"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.sqs.QueueDLQ import QueueDLQ


@pytest.fixture(scope="module")
def rule():
    rule = QueueDLQ()
    yield rule


@pytest.fixture
def template():
    return {
        "Parameters": {"FifoQueue": {"Type": "Boolean"}},
        "Conditions": {"UseFifoQueue": {"Fn::Equals": [{"Ref": "FifoQueue"}, "true"]}},
        "Resources": {
            "CustomResource": {"Type": "AWS::CloudFormation::CustomResource"},
            "FifoQueue": {
                "Type": "AWS::SQS::Queue",
                "Properties": {
                    "FifoQueue": True,
                },
            },
            "StandardQueue1": {
                "Type": "AWS::SQS::Queue",
                "Properties": {},
            },
            "StandardQueue2": {
                "Type": "AWS::SQS::Queue",
                "Properties": {
                    "FifoQueue": "false",
                },
            },
        },
    }


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "FifoQueue": True,
                "RedrivePolicy": {
                    "deadLetterTargetArn": {"Fn::GetAtt": "FifoQueue.Arn"}
                },
            },
            [],
        ),
        (
            {
                "FifoQueue": True,
                "RedrivePolicy": {
                    "deadLetterTargetArn": {"Fn::GetAtt": "CustomResource.Arn"}
                },
            },
            [],
        ),
        (
            {
                "RedrivePolicy": {
                    "deadLetterTargetArn": {"Fn::GetAtt": "StandardQueue1.Arn"}
                }
            },
            [],
        ),
        (
            {
                "RedrivePolicy": {
                    "deadLetterTargetArn": {"Fn::GetAtt": "StandardQueue2.Arn"}
                }
            },
            [],
        ),
        (
            {"RedrivePolicy": {"deadLetterTargetArn": {"Fn::GetAtt": []}}},
            [],
        ),
        (
            {
                "RedrivePolicy": {
                    "deadLetterTargetArn": {"Fn::GetAtt": "CustomResource.Arn"}
                }
            },
            [],
        ),
        (
            {
                "FifoQueue": {
                    "Fn::If": [
                        "UseFifoQueue",
                        True,
                        False,
                    ]
                },
                "RedrivePolicy": {
                    "deadLetterTargetArn": {
                        "Fn::If": [
                            "UseFifoQueue",
                            {"Fn::GetAtt": "FifoQueue.Arn"},
                            {"Fn::GetAtt": "StandardQueue1.Arn"},
                        ]
                    }
                },
            },
            [],
        ),
        (
            {
                "FifoQueue": True,
                "RedrivePolicy": {
                    "deadLetterTargetArn": {"Fn::GetAtt": "StandardQueue1.Arn"}
                },
            },
            [
                ValidationError(
                    (
                        "Source queue type 'FIFO' does not match "
                        "destination queue type 'standard'"
                    ),
                    rule=QueueDLQ(),
                    path=deque(["RedrivePolicy", "deadLetterTargetArn"]),
                )
            ],
        ),
        (
            {
                "FifoQueue": True,
                "RedrivePolicy": {
                    "deadLetterTargetArn": {"Fn::GetAtt": "StandardQueue2.Arn"}
                },
            },
            [
                ValidationError(
                    (
                        "Source queue type 'FIFO' does not match "
                        "destination queue type 'standard'"
                    ),
                    rule=QueueDLQ(),
                    path=deque(["RedrivePolicy", "deadLetterTargetArn"]),
                )
            ],
        ),
        (
            {
                "FifoQueue": "false",
                "RedrivePolicy": {
                    "deadLetterTargetArn": {"Fn::GetAtt": "FifoQueue.Arn"}
                },
            },
            [
                ValidationError(
                    (
                        "Source queue type 'standard' does not "
                        "match destination queue type 'FIFO'"
                    ),
                    rule=QueueDLQ(),
                    path=deque(["RedrivePolicy", "deadLetterTargetArn"]),
                )
            ],
        ),
        (
            {
                "FifoQueue": {
                    "Fn::If": [
                        "UseFifoQueue",
                        True,
                        False,
                    ]
                },
                "RedrivePolicy": {
                    "deadLetterTargetArn": {
                        "Fn::If": [
                            "UseFifoQueue",
                            {"Fn::GetAtt": "StandardQueue1.Arn"},
                            {"Fn::GetAtt": "FifoQueue.Arn"},
                        ]
                    }
                },
            },
            [
                ValidationError(
                    (
                        "Source queue type 'FIFO' does not match "
                        "destination queue type 'standard'"
                    ),
                    rule=QueueDLQ(),
                    path=deque(["RedrivePolicy", "deadLetterTargetArn", "Fn::If", 1]),
                ),
                ValidationError(
                    (
                        "Source queue type 'standard' does "
                        "not match destination queue type 'FIFO'"
                    ),
                    rule=QueueDLQ(),
                    path=deque(["RedrivePolicy", "deadLetterTargetArn", "Fn::If", 2]),
                ),
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.templates.LimitSize import LimitSize  # pylint: disable=E0401


def write_limit_test_templates():
    with open("test/fixtures/templates/bad/limit_numbers.yaml", "w") as f:
        f.write("Parameters:\n")
        for n in range(1, 202):
            f.write(
                "  Parameter"
                + str(n)
                + """:
    Type: String\n"""
            )
        f.write("Resources:\n")
        for n in range(1, 502):
            f.write(
                "  Resource"
                + str(n)
                + """:
    Type: AWS::SNS::Topic\n"""
            )
        f.write("Outputs:\n")
        for n in range(1, 202):
            f.write(
                "  Output"
                + str(n)
                + """:
    Value: !Ref Parameter1\n"""
            )
        f.write("Mappings:\n")
        for n in range(1, 202):
            f.write(
                "  Mapping"
                + str(n)
                + """:
    Key:
      Key: Value\n"""
            )
        for n in range(1, 202):
            f.write("      Key" + str(n) + ": Value\n")

    with open("test/fixtures/templates/bad/limit_size.yaml", "w") as f:
        f.write(
            """Description: "Lorem ipsum dolor sit amet, consectetur adipiscing elit. In aliquet justo nibh, a venenatis justo suscipit et. Sed vulputate est non vulputate cursus. In et mollis nunc. Ut semper justo nec odio dignissim semper. Sed interdum elementum ante. Phasellus bibendum mattis ultrices. Nullam varius dui mi, ac fermentum ante sodales nec. Mauris id libero id turpis sollicitudin imperdiet. Praesent volutpat, elit ut malesuada ultricies, magna velit ullamcorper leo, sagittis pulvinar elit erat faucibus felis.
Morbi sagittis pretium ligula in tristique. Suspendisse odio lectus, condimentum eget egestas pharetra, elementum sit amet sapien. Nullam ipsum eros, ullamcorper ut mi nec, maximus sagittis lacus. In laoreet quam sit amet ante dictum efficitur. Praesent pellentesque purus sit amet nisl scelerisque feugiat. Phasellus feugiat, ex posuere pharetra eleifend, ipsum libero viverra est, sit amet sodales ipsum urna in quam. Donec sit amet pharetra nibh. Nulla eu fermentum leo, a venenatis urna. Nam ac sagittis magna volutpat."
Resources:\n"""
        )
        for n in range(1, 300):
            # ruff: noqa: E501
            f.write(
                "  Resource"
                + str(n)
                + """:
    Type: AWS::EC2::Instance
    Metadata:
      Comment: Install a Magento web serve
      AWS::CloudFormation::Authentication:
        S3AccessCreds:
          type: S3
          roleName: NewIamInstanceRole
      AWS::CloudFormation::Init:
        config:
          packages:
            yum:
              nfs-utils: []
              awslogs: []
          files:
            /etc/awslogs/awslogs.conf:
              content: !Sub |
                [general]
                state_file= /var/log/agent-state
                [/home/ec2-user/misc/install]
                file = /var/log/cloud-init-output.log
                log_group_name = MagentoMainLogGroup
                log_stream_name = magento/install.log
              mode: '000400'
              owner: root
              group: root
            /etc/awslogs/awscli.conf:
              content: !Sub |
                [plugins]
                cwlogs = cwlogs
                [default]
                region = ${AWS::Region}
              mode: '000400'
              owner: root
              group: root
            /etc/cfn/cfn-hup.conf:
              content: !Sub |
                [main]
                stack=${AWS::StackId}
                region=${AWS::Region}
              mode: '000400'
              owner: root
              group: root
            /etc/cfn/hooks.d/cfn-auto-reloader.conf:
              content: !Sub |
                [cfn-auto-reloader-hook]
                triggers=post.update
                path=Resources.LaunchConfig.Metadata.AWS::CloudFormation::Init
                action=/opt/aws/bin/cfn-init -v --stack ${AWS::StackName} --resource LaunchConfig --region ${AWS::Region}
                runas=root
            /install_magento.sh:
              source: !Sub https://quickstart-reference.s3.amazonaws.com/magento/latest/scripts/install_magento.sh
              mode: '000644'
              owner: root
              group: root
              authentication: S3AccessCreds
            /configure_magento.sh:
              source: !Sub https://quickstart-reference.s3.amazonaws.com/magento/latest/scripts/configure_magento.sh
              mode: '000644'
              owner: root
              group: root
              authentication: S3AccessCreds
            /home/ec2-user/magento.tar.gz:
              source: 'magento.tar.gz'
              mode: '000644'
              owner: root
              group: root
              authentication: S3AccessCreds
          services:
            sysvinit:
              cfn-hup:
                enabled: 'true'
                ensureRunning: 'true'
                files:
                - /etc/cfn/cfn-hup.conf
                - /etc/cfn/hooks.d/cfn-auto-reloader.conf
              awslogs:
                enabled: 'true'
                ensureRunning: 'true'
                files:
                - /etc/awslogs/awslogs.conf
                packages:
                  yum:
                  - awslogs
    Properties:
      ImageId: ami-ami
      InstanceType: t2.micro
      UserData:
        !Base64
          Fn::Sub: |
            #!/bin/bash
            /opt/aws/bin/cfn-init -v --stack ${AWS::StackName} --resource MagentoAMI  --region ${AWS::Region}
            chmod a+x install_magento.sh
            ./install_magento.sh /tmp/params.txt
      Tags:
      - Key: Application
        Value: AWS Quick Start (Magento)
      - Key: Name
        Value: MagentoAMI (AWS Quick Start)
"""
            )


class TestTemplateLimitSize(BaseRuleTestCase):
    """Test template limit size"""

    def setUp(self):
        """Setup"""
        super(TestTemplateLimitSize, self).setUp()
        self.collection.register(LimitSize())
        write_limit_test_templates()

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative("test/fixtures/templates/bad/limit_size.yaml", 1)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.jsonschema._utils import Unset
from cfnlint.rules import CloudFormationLintRule


class CfnLint(CloudFormationLintRule):
    id = "E1101"
    shortdesc = "Validate an item against additional checks"
    description = "Use supplemental logic to validate properties against"
    tags = []

    # pylint: disable=unused-argument
    def cfnLint(self, validator, keywords, instance, schema):
        validator = validator.evolve(
            function_filter=validator.function_filter.evolve(
                add_cfn_lint_keyword=False,
            )
        )

        for keyword in keywords:
            for rule in self.child_rules.values():
                if rule is None:
                    continue
                if not rule.id:
                    continue

                for rule_keyword in rule.keywords:
                    if rule_keyword == keyword:
                        for err in rule.validate(validator, keyword, instance, schema):
                            if err.rule is None:
                                if isinstance(err.validator, Unset):
                                    err.rule = rule
                            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import itertools
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Iterator

from sympy import Not, Or
from sympy.logic.boolalg import BooleanFunction
from sympy.logic.inference import satisfiable

from cfnlint.conditions._utils import get_hash
from cfnlint.context.conditions._condition import Condition
from cfnlint.context.conditions._utils import (
    build_instance_from_scenario,
    get_conditions_from_property,
)
from cfnlint.context.conditions.exceptions import Unsatisfiable

if TYPE_CHECKING:
    from cfnlint.context.context import Context, Parameter


@dataclass(frozen=True)
class Conditions:
    # Template level condition management
    conditions: dict[str, Condition] = field(init=True, default_factory=dict)
    cnf: BooleanFunction | None = field(init=True, default=None)
    _max_scenarios: int = field(init=False, default=128)

    @classmethod
    def create_from_instance(
        cls, conditions: Any, rules: dict[str, dict], parameters: dict[str, "Parameter"]
    ) -> "Conditions":
        obj: dict[str, Condition] = {}
        if not isinstance(conditions, dict):
            raise ValueError("Conditions must be a object")
        for k, v in conditions.items():
            try:
                other_conditions = conditions.copy()
                del other_conditions[k]
                obj[k] = Condition.create_from_instance(v, other_conditions)
            except ValueError:
                # this is a default condition so we can keep the name but it will
                # not associate with another condition and will always be true/false
                obj[k] = Condition.create_from_instance(
                    {"Fn::Equals": [None, None]}, conditions
                )

        cnf = None
        for p_k, p_v in parameters.items():

            if not p_v.allowed_values:
                continue
            allowed_values = p_v.allowed_values.copy()
            equals_cnfs = []
            for _, c_v in obj.items():
                for i in c_v.equals:
                    if i.right.hash == get_hash({"Ref": p_k}):
                        if not isinstance(i.left.instance, str):
                            continue
                        equals_cnfs.append(i.cnf)
                        if i.left.instance in allowed_values:
                            allowed_values.remove(i.left.instance)

            if not allowed_values:
                if cnf is None:
                    cnf = Or(*equals_cnfs)
                else:
                    cnf = cnf & Or(*equals_cnfs)

        return cls(conditions=obj, cnf=cnf)

    def evolve(self, status: dict[str, bool]) -> "Conditions":
        cls = self.__class__

        conditions: dict[str, Condition] = {}
        cnf = self.cnf
        for condition, value in self.conditions.items():
            s = status.get(condition, value.status)
            try:
                conditions[condition] = value.evolve(status=s)
                if s is not None:
                    if cnf:
                        cnf = (
                            cnf & conditions[condition].cnf
                            if s
                            else cnf & Not(conditions[condition].cnf)
                        )
                    else:
                        cnf = (
                            conditions[condition].cnf
                            if s
                            else Not(conditions[condition].cnf)
                        )
            except ValueError as e:
                raise Unsatisfiable(
                    new_status=status,
                    current_status=self.status,
                ) from e

        if not satisfiable(cnf):
            raise Unsatisfiable(
                new_status=status,
                current_status=self.status,
            )

        return cls(
            conditions=conditions,
            cnf=cnf,
        )

    def _build_conditions(self, conditions: set[str]) -> Iterator["Conditions"]:
        scenarios_attempted = 0
        for product in itertools.product([True, False], repeat=len(conditions)):
            params = dict(zip(conditions, product))
            try:
                yield self.evolve(params)
            except Unsatisfiable:
                pass

            scenarios_attempted += 1
            # On occassions people will use a lot of non-related conditions
            # this is fail safe to limit the maximum number of responses
            if scenarios_attempted >= self._max_scenarios:
                return

    def evolve_from_instance(
        self, instance: Any, context: "Context"
    ) -> Iterator[tuple[Any, "Conditions"]]:

        conditions = get_conditions_from_property(instance)

        for scenario in self._build_conditions(conditions):
            yield build_instance_from_scenario(
                instance,
                scenario.status,
                is_root=True,
                context=context,
            ), scenario

    @property
    def status(self) -> dict[str, bool]:
        obj = {}
        for name, c in self.conditions.items():
            if c.status is not None:
                obj[name] = c.status

        return obj



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.rules.resources.properties.AvailabilityZone import (
    AvailabilityZone,  # pylint: disable=E0401
)


@pytest.fixture(scope="module")
def rule():
    rule = AvailabilityZone()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MySqs": {
                "Type": "AWS::SQS::Queue",
            },
            "CDK": {
                "Type": "AWS::CDK::Metadata",
            },
        }
    }


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid hardcoded string because CDK",
            "us-east-1a",
            {
                "path": deque(["Resources", "MySqs", "Properties"]),
            },
            [],
        ),
    ],
    indirect=["path"],
)
def test_validate(name, instance, path, expected, rule, validator):
    errors = list(rule.validate(validator, False, instance, {}))
    # we use error counts in this one as the instance types are
    # always changing so we aren't going to hold ourselves up by that
    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Helpers for loading resources, managing specs, constants, etc.

Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from typing import Any, Callable

from cfnlint.jsonschema.exceptions import ValidationError

_keyword = "format"

LOGGER = logging.getLogger(__name__)


def _any_of(
    source: str,
    any_ofs: list[dict[str, Any]],
) -> tuple[bool, list[str]]:
    """Compare the source schema with the destination schema and
    return a ValidationError if they don't match.

    Args:
        source (str): The source schema to compare.
        destination (list[dict[str, Any]]): The anyOf schemas

    Returns:
        tuple[bool, list[str]]: A bool value if format matches and a
                                list of format values if they don't match.
    """
    formats = []
    for subschema in any_ofs:
        if _keyword in subschema:
            if source == subschema[_keyword]:
                return True, []

            formats.append(subschema[_keyword])

    return False, formats


_schema_composition: dict[str, Callable] = {
    "anyOf": _any_of,
}


def _backwards_compatibility(format: Any) -> Any:
    if format == "AWS::EC2::SecurityGroup.GroupId":
        LOGGER.warning(
            f"{format!r} is deprecated. Use 'AWS::EC2::SecurityGroup.Id' instead"
        )
        return "AWS::EC2::SecurityGroup.Id"
    return format


def _compare_schemas(
    source: str,
    destination: dict[str, Any],
) -> tuple[bool, list[str]]:
    """Compare two schemas and return a ValidationError if they don't match.

    Args:
        source (str): The source schema to compare.
        destination (dict[str, Any]): The destination schema to compare against.

    Returns:
        tuple[bool, list[str]]: True/False if the format keyword matches
    """

    dest_f = _backwards_compatibility(destination.get(_keyword))
    if dest_f == source:
        return True, [dest_f]  # Nothing else matters to be on the safe side

    all_formats = [dest_f] if dest_f else []
    for composition_key in _schema_composition:
        if composition_key in destination:
            match, formats = _schema_composition[composition_key](
                source, destination[composition_key]
            )
            if match:
                return True, []
            all_formats.extend(formats)

    return False, all_formats


def compare_schemas(
    source: dict[str, Any],
    destination: dict[str, Any],
) -> ValidationError | None:
    """Compare two schemas and return a ValidationError if they don't match.

    Args:
        source (dict[str, Any]): The source schema to compare.
        destination (dict[str, Any]): The destination schema to compare against.

    Returns:
        ValidationError: A ValidationError if the schemas don't match, otherwise None.
    """

    f = _backwards_compatibility(source.get(_keyword))

    if f is None:
        return None

    match, formats = _compare_schemas(f, destination)
    if match:
        return None

    if formats:
        return ValidationError(
            f"{f!r} format is incompatible with formats {formats!r}",
            schema={"format": f},
            instance=formats,
        )

    return ValidationError(
        f"{f!r} format is incompatible",
        schema={"format": f},
        instance=None,
    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.s3.BucketTieringConfiguration import (
    BucketTieringConfiguration,
)


@pytest.fixture(scope="module")
def rule():
    rule = BucketTieringConfiguration()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {},
            [],
        ),
        (
            {"AccessTier": "ARCHIVE_ACCESS"},
            [],
        ),
        (
            {
                "AccessTier": "ARCHIVE_ACCESS",
                "Days": "90",
            },
            [],
        ),
        (
            {
                "AccessTier": "ARCHIVE_ACCESS",
                "Days": "a",
            },
            [],
        ),
        (
            {
                "AccessTier": "ARCHIVE_ACCESS",
                "Days": 1,
            },
            [
                ValidationError(
                    ("1 is less than the minimum of 90"),
                    path=deque(["Days"]),
                    validator="minimum",
                    schema_path=deque(["minimum"]),
                    rule=BucketTieringConfiguration(),
                )
            ],
        ),
        (
            {
                "AccessTier": "ARCHIVE_ACCESS",
                "Days": 99999,
            },
            [
                ValidationError(
                    ("99999 is greater than the maximum of 730"),
                    path=deque(["Days"]),
                    validator="maximum",
                    schema_path=deque(["maximum"]),
                    rule=BucketTieringConfiguration(),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import regex as re

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class ResourcePolicyResourceArn(CfnLintKeyword):

    id = "E3514"
    shortdesc = "Validate IAM resource policy resource ARNs"
    description = "Validates an IAM resource policy has a compliant resource ARN"
    source_url = (
        "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_identifiers.html"
    )
    tags = ["parameters", "iam"]

    def __init__(self) -> None:
        super().__init__(
            [
                "AWS::IAM::Policy/Properties/PolicyDocument/Statement/Resource",
            ]
        )

    # pylint: disable=unused-argument
    def validate(
        self, validator: Validator, aZ: Any, arn: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(arn, "string"):
            return

        if not len(validator.context.path.cfn_path) >= 2:
            return

        patrn: str = (
            "^(arn:aws[A-Za-z\\-]*?:[^:]+:[^:]*(:(?:\\d{12}|\\*|aws)?:.+|)|\\*)$"
        )
        if validator.context.path.cfn_path[1] == "AWS::S3::BucketPolicy":
            patrn = "^arn:aws[A-Za-z\\-]*?:[^:]+:[^:]*(:(?:\\d{12}|\\*|aws)?:.+|)$"

        if not re.match(patrn, arn):
            yield ValidationError(
                f"{arn!r} does not match {patrn!r}",
                rule=self,
            )



import json
import logging
from test.testlib.testcase import BaseTestCase
from unittest.mock import mock_open, patch

from cfnlint.schema._patch import SchemaPatch
from cfnlint.schema._patch import patch as schema_patch
from cfnlint.schema._patch import reset
from cfnlint.schema.manager import ProviderSchemaManager

LOGGER = logging.getLogger("cfnlint.schema.manager")
LOGGER.disabled = True


class TestSchemaPatch(BaseTestCase):
    """Test the SchemaPatch class"""

    def test_from_dict(self):
        """Test creating a SchemaPatch from a dictionary"""
        test_dict = {
            "IncludeResourceTypes": ["AWS::EC2::*"],
            "ExcludeResourceTypes": ["AWS::EC2::VPC"],
            "Patches": {
                "AWS::EC2::Instance": [
                    {"op": "add", "path": "/properties/test", "value": "value"}
                ]
            },
        }

        patch = SchemaPatch.from_dict(test_dict)

        self.assertEqual(patch.included_resource_types, ["AWS::EC2::*"])
        self.assertEqual(patch.excluded_resource_types, ["AWS::EC2::VPC"])
        self.assertEqual(
            patch.patches,
            {
                "AWS::EC2::Instance": [
                    {"op": "add", "path": "/properties/test", "value": "value"}
                ]
            },
        )

    def test_from_dict_empty(self):
        """Test creating a SchemaPatch from an empty dictionary"""
        test_dict = {}

        patch = SchemaPatch.from_dict(test_dict)

        self.assertEqual(patch.included_resource_types, [])
        self.assertEqual(patch.excluded_resource_types, [])
        self.assertEqual(patch.patches, {})


class TestReset(BaseTestCase):
    """Test the reset function"""

    @patch("cfnlint.schema._patch.PROVIDER_SCHEMA_MANAGER")
    @patch("cfnlint.schema._patch.OTHER_SCHEMA_MANAGER")
    def test_reset(self, mock_other_manager, mock_provider_manager):
        """Test that reset calls reset on both managers"""
        reset()

        mock_provider_manager.reset.assert_called_once()
        mock_other_manager.reset.assert_called_once()


class TestManagerPatch(BaseTestCase):
    """Used for Testing Resource Schemas"""

    def setUp(self) -> None:
        super().setUp()

        self.manager = ProviderSchemaManager()
        self.schemas = dict.fromkeys(["aws-lambda-codesigningconfig"])
        for resource in self.schemas:
            with open(f"test/fixtures/registry/schemas/{resource}.json") as fh:
                self.schemas[resource] = fh.read()

    def test_patch_file_not_found_error(self):
        with patch("builtins.open", mock_open()) as mock_builtin_open:
            err = FileNotFoundError()
            err.errno = 2
            mock_builtin_open.side_effect = [err]
            with self.assertRaises(SystemExit) as mock_exit:
                schema_patch("bad", regions=["us-east-1"])
                self.assertEqual(mock_exit.type, SystemExit)
                self.assertEqual(mock_exit.value.code == 1)

    def test_patch_file_is_dir(self):
        with patch("builtins.open", mock_open()) as mock_builtin_open:
            err = IOError()
            err.errno = 21
            mock_builtin_open.side_effect = [err]
            with self.assertRaises(SystemExit) as mock_exit:
                schema_patch("bad", regions=["us-east-1"])
                self.assertEqual(mock_exit.type, SystemExit)
                self.assertEqual(mock_exit.value.code == 1)

    def test_patch_permission_error(self):
        with patch("builtins.open", mock_open()) as mock_builtin_open:
            err = PermissionError()
            err.errno = 13
            mock_builtin_open.side_effect = [err]
            with self.assertRaises(SystemExit) as mock_exit:
                schema_patch("bad", regions=["us-east-1"])
                self.assertEqual(mock_exit.type, SystemExit)
                self.assertEqual(mock_exit.value.code == 1)

    def test_patch_value_error(self):
        with patch("builtins.open", mock_open()) as mock_builtin_open:
            err = ValueError()
            mock_builtin_open.side_effect = [err]
            with self.assertRaises(SystemExit) as mock_exit:
                schema_patch("bad", regions=["us-east-1"])
                self.assertEqual(mock_exit.type, SystemExit)
                self.assertEqual(mock_exit.value.code == 1)

    @patch("cfnlint.schema._patch.PROVIDER_SCHEMA_MANAGER")
    @patch("cfnlint.schema._patch.OTHER_SCHEMA_MANAGER")
    def test_patch_success(self, mock_other_manager, mock_provider_manager):
        """Test successful patching"""
        test_patch_data = {
            "IncludeResourceTypes": ["AWS::EC2::*"],
            "ExcludeResourceTypes": ["AWS::EC2::VPC"],
            "Patches": {
                "AWS::EC2::Instance": [
                    {"op": "add", "path": "/properties/test", "value": "value"}
                ]
            },
        }

        with patch("builtins.open", mock_open(read_data=json.dumps(test_patch_data))):
            schema_patch("test_file.json", regions=["us-east-1", "us-west-2"])

            # Check that provider manager patch was called for each region
            self.assertEqual(mock_provider_manager.patch.call_count, 2)

            # Check that other manager patch was called once (with the last region)
            mock_other_manager.patch.assert_called_once()



__all__ = ["Context", "create_context_for_template"]

from cfnlint.context.conditions.exceptions import Unsatisfiable
from cfnlint.context.context import Context, Path, Resource, create_context_for_template



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import Validator
from cfnlint.rules import CloudFormationLintRule


class Exists(CloudFormationLintRule):
    """Check if used Conditions are defined"""

    id = "E8002"
    shortdesc = "Check if the referenced Conditions are defined"
    description = (
        "Making sure the used conditions are actually defined in the Conditions section"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/conditions-section-structure.html"
    tags = ["conditions"]

    def cfncondition(self, validator: Validator, conditions, instance: Any, schema):
        if not validator.is_type(instance, "string"):
            return
        for err in validator.descend(
            instance, {"enum": list(validator.context.conditions.conditions.keys())}
        ):
            err.rule = self
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context._mappings import Mappings
from cfnlint.context.context import (
    Context,
    Parameter,
    Path,
    Resource,
    create_context_for_template,
)
from cfnlint.jsonschema import ValidationError
from cfnlint.jsonschema.validators import CfnTemplateValidator
from cfnlint.template import Template


def _resolve(name, instance, expected_results, **kwargs):
    validator = CfnTemplateValidator().evolve(**kwargs)

    resolutions = list(validator.resolve_value(instance))

    assert len(resolutions) == len(
        expected_results
    ), f"{name!r} got {len(resolutions)!r}"

    for i, (instance, v, errors) in enumerate(resolutions):
        assert instance == expected_results[i][0], f"{name!r} got {instance!r}"
        assert errors == expected_results[i][2], f"{name!r} got {errors!r}"
        if not errors:
            expected_context = expected_results[i][1]
            _, expected_context = list(expected_context.ref_value("AWS::Region"))[0]
            assert (
                v.context.path == expected_context.path
            ), f"{name!r} got {v.context.path!r}"
            assert (
                v.context.ref_values == expected_context.ref_values
            ), f"{name!r} got {v.context.ref_values!r}"


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Valid Ref with a single instance",
            {"Ref": {"Ref": "MyResource"}},
            [],
        ),
    ],
)
def test_resolvers_ref(name, instance, response):
    _resolve(name, instance, response)


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Invalid Ref with an array",
            {"Ref": []},
            [],
        ),
        (
            "Invalid Ref with an invalid object",
            {"Ref": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid Join with an invalid type",
            {"Fn::Join": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid Join with an invalid array length",
            {"Fn::Join": ["a", "b", "c"]},
            [],
        ),
        (
            "Invalid Join with an invalid type for first element",
            {"Fn::Join": [["a"], "b"]},
            [],
        ),
        (
            "Invalid Join with an invalid type for second element",
            {"Fn::Join": ["a", "b"]},
            [],
        ),
        (
            "Invalid Join with an invalid type for second element item",
            {"Fn::Join": ["/", [["a"], "b"]]},
            [],
        ),
        (
            "Invalid Select with an invalid type",
            {"Fn::Select": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid Select with an invalid array length",
            {"Fn::Select": ["a", "b", "c"]},
            [],
        ),
        (
            "Invalid Select with an invalid type for second element",
            {"Fn::Select": ["a", "b"]},
            [],
        ),
        (
            "Invalid Select with an invalid type for second element",
            {"Fn::Select": [0, "b"]},
            [],
        ),
        (
            "Invalid Split with an invalid type",
            {"Fn::Split": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid Split with an invalid array length",
            {"Fn::Split": ["a", "b", "c"]},
            [],
        ),
        (
            "Invalid Split with an invalid type for first element",
            {"Fn::Split": [0, "b"]},
            [],
        ),
        (
            "Invalid Split with an invalid type for second element",
            {"Fn::Split": ["a", ["b"]]},
            [],
        ),
        (
            "Invalid GetAZs with an invalid type ",
            {"Fn::GetAZs": {"a": "b"}},
            [],
        ),
        (
            "Invalid GetAZs with an invalid type ",
            {"Fn::GetAZs": ["a", "b"]},
            [],
        ),
        (
            "Invalid GetAZs with an invalid region",
            {"Fn::GetAZs": "foo"},
            [],
        ),
        (
            "Invalid FindInMap with an invalid type",
            {"Fn::FindInMap": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid FindInMap with an invalid length",
            {"Fn::FindInMap": ["foo", "bar"]},
            [],
        ),
        (
            "Invalid FindInMap with an invalid type for first element",
            {"Fn::FindInMap": [["foo"], "bar", "value"]},
            [],
        ),
        (
            "Invalid FindInMap with an invalid type for second element",
            {"Fn::FindInMap": ["foo", ["bar"], "value"]},
            [],
        ),
        (
            "Invalid FindInMap with an invalid type for third element",
            {"Fn::FindInMap": ["foo", "first", ["value"]]},
            [],
        ),
        (
            "Invalid FnSub with an invalid type",
            {"Fn::Sub": {"foo": "bar"}},
            [],
        ),
        (
            "Invalid FnSub with an array of wrong length",
            {"Fn::Sub": ["foo", "bar", "value"]},
            [],
        ),
        (
            "Invalid FnSub with the wrong type for element one",
            {"Fn::Sub": [["foo"], {"foo": "bar"}]},
            [],
        ),
        (
            "Invalid FnSub with the wrong type for element two",
            {"Fn::Sub": ["foo", ["bar"]]},
            [],
        ),
    ],
)
def test_invalid_functions(name, instance, response):
    context = Context(
        mappings=Mappings.create_from_dict({"foo": {"first": {"second": "bar"}}})
    )

    _resolve(name, instance, response, context=context)


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Valid Join with an empty type",
            {"Fn::Join": [".", []]},
            [],
        ),
        (
            "Valid Join with a list of strings",
            {"Fn::Join": [".", ["a", "b", "c"]]},
            [("a.b.c", Context(), None)],
        ),
        (
            "Valid GetAZs with empty string",
            {"Fn::GetAZs": ""},
            [
                (
                    [
                        "us-east-1a",
                        "us-east-1b",
                        "us-east-1c",
                        "us-east-1d",
                        "us-east-1e",
                        "us-east-1f",
                    ],
                    Context(),
                    None,
                )
            ],
        ),
        (
            "Valid FindInMap with bad keys and a default value",
            {"Fn::FindInMap": ["foo", "bar", "value", {"DefaultValue": "default"}]},
            [
                (
                    "default",
                    Context(path=Path(value_path=deque([4, "DefaultValue"]))),
                    None,
                )
            ],
        ),
        (
            "Valid FindInMap with valid keys and a default value",
            {"Fn::FindInMap": ["foo", "first", "second", {"DefaultValue": "default"}]},
            [
                (
                    "default",
                    Context(path=Path(value_path=deque([4, "DefaultValue"]))),
                    None,
                ),
                (
                    "bar",
                    Context(
                        path=Path(
                            value_path=deque(["Mappings", "foo", "first", "second"])
                        )
                    ),
                    None,
                ),
            ],
        ),
        (
            "Valid FindInMap with a bad mapping",
            {"Fn::FindInMap": ["bar", "first", "second"]},
            [
                (
                    None,
                    Context(),
                    ValidationError(
                        (
                            "'bar' is not one of ['foo', "
                            "'transformFirstKey', 'transformSecondKey', "
                            "'integers', 'accounts', 'environments']"
                        ),
                        path=deque(["Fn::FindInMap", 0]),
                    ),
                )
            ],
        ),
        (
            "Valid FindInMap with a bad mapping and default",
            {"Fn::FindInMap": ["bar", "first", "second", {"DefaultValue": "default"}]},
            [
                (
                    "default",
                    Context(path=Path(value_path=deque([4, "DefaultValue"]))),
                    None,
                )
            ],
        ),
        (
            "Valid FindInMap with a bad mapping and aws no value",
            {
                "Fn::FindInMap": [
                    "bar",
                    "first",
                    "second",
                    {"DefaultValue": {"Ref": "AWS::NoValue"}},
                ]
            },
            [],
        ),
        (
            "Valid FindInMap with a bad top key",
            {"Fn::FindInMap": ["foo", "second", "first"]},
            [
                (
                    None,
                    Context(),
                    ValidationError(
                        ("'second' is not one of ['first'] for " "mapping 'foo'"),
                        path=deque(["Fn::FindInMap", 1]),
                    ),
                )
            ],
        ),
        (
            "Valid FindInMap with a bad top key and default",
            {"Fn::FindInMap": ["foo", "second", "first", {"DefaultValue": "default"}]},
            [
                (
                    "default",
                    Context(path=Path(value_path=deque([4, "DefaultValue"]))),
                    None,
                )
            ],
        ),
        (
            "Valid FindInMap with a map name that is a Ref to pseudo param",
            {"Fn::FindInMap": [{"Ref": "AWS::StackName"}, "first", "second"]},
            [],
        ),
        (
            "Valid FindInMap with an top level key that is a Ref to an account",
            {"Fn::FindInMap": ["accounts", {"Ref": "AWS::AccountId"}, "dev"]},
            [
                (
                    "bar",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "accounts", "123456789012", "dev"]
                            )
                        ),
                    ),
                    None,
                )
            ],
        ),
        (
            (
                "Valid FindInMap with an top level key "
                "that is a Ref to non account non region"
            ),
            {"Fn::FindInMap": ["accounts", {"Ref": "AWS::StackName"}, "dev"]},
            [
                (
                    "bar",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "accounts", "123456789012", "dev"]
                            )
                        ),
                    ),
                    None,
                )
            ],
        ),
        (
            "Valid FindInMap with an top level key that is a Ref to a region",
            {"Fn::FindInMap": ["accounts", {"Ref": "AWS::Region"}, "bar"]},
            [
                (
                    "foo",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "accounts", "us-east-1", "bar"]
                            )
                        ),
                    ),
                    None,
                )
            ],
        ),
        (
            "Invalid FindInMap with an top level key that is a Ref to an account",
            {"Fn::FindInMap": ["accounts", {"Ref": "AWS::AccountId"}, "bar"]},
            [
                (
                    None,
                    Context(),
                    ValidationError(
                        (
                            "'bar' is not a second level key "
                            "when {'Ref': 'AWS::AccountId'} is "
                            "resolved for mapping 'accounts'"
                        ),
                        path=deque(["Fn::FindInMap", 2]),
                    ),
                )
            ],
        ),
        (
            "Invalid FindInMap with an top level key that is a Ref to pseudo param",
            {"Fn::FindInMap": ["foo", {"Ref": "AWS::AccountId"}, "second"]},
            [
                (
                    None,
                    Context(),
                    ValidationError(
                        (
                            "{'Ref': 'AWS::AccountId'} is not a "
                            "first level key for mapping 'foo'"
                        ),
                        path=deque(["Fn::FindInMap", 1]),
                    ),
                )
            ],
        ),
        (
            "Valid FindInMap with a second level key that is a Ref to pseudo param",
            {"Fn::FindInMap": ["foo", "first", {"Ref": "AWS::AccountId"}]},
            [],
        ),
        (
            "Valid FindInMap with a bad third key",
            {"Fn::FindInMap": ["foo", "first", "third"]},
            [
                (
                    None,
                    Context(),
                    ValidationError(
                        (
                            "'third' is not one of ['second'] for "
                            "mapping 'foo' and key 'first'"
                        ),
                        path=deque(["Fn::FindInMap", 2]),
                    ),
                )
            ],
        ),
        (
            "Valid FindInMap with integer types",
            {"Fn::FindInMap": ["integers", 1, 2]},
            [
                (
                    "Value",
                    Context(
                        path=Path(value_path=deque(["Mappings", "integers", "1", "2"]))
                    ),
                    None,
                )
            ],
        ),
        (
            (
                "Valid FindInMap with a Ref to a parameter "
                "with allowed values for top level key"
            ),
            {"Fn::FindInMap": ["environments", {"Ref": "Environment"}, "foo"]},
            [
                (
                    "one",
                    Context(
                        path=Path(
                            value_path=deque(["Mappings", "environments", "dev", "foo"])
                        ),
                        ref_values={"Environment": "dev"},
                    ),
                    None,
                ),
                (
                    "two",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "test", "foo"]
                            ),
                        ),
                        ref_values={"Environment": "test"},
                    ),
                    None,
                ),
            ],
        ),
        (
            "Valid FindInMap with a Ref to a parameter for top level key",
            {"Fn::FindInMap": ["environments", {"Ref": "RandomString"}, "foo"]},
            [],
        ),
        (
            (
                "Valid FindInMap with a Ref to accounts for top level "
                "key Ref to a ramom string for second level key"
            ),
            {
                "Fn::FindInMap": [
                    "accounts",
                    {"Ref": "AWS::AccountId"},
                    {"Ref": "RandomString"},
                ]
            },
            [],
        ),
        (
            (
                "Valid FindInMap with a Ref to accounts for top level "
                "key Ref to an allowed value parameter for second level key"
            ),
            {
                "Fn::FindInMap": [
                    "accounts",
                    {"Ref": "AWS::AccountId"},
                    {"Ref": "Environment"},
                ]
            },
            [],
        ),
        (
            (
                "Valid FindInMap with a Ref to a parameter "
                "with allowed values for second level key"
            ),
            {"Fn::FindInMap": ["environments", "lion", {"Ref": "Environment"}]},
            [
                (
                    "one",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "dev"]
                            )
                        ),
                        ref_values={"Environment": "dev"},
                    ),
                    None,
                ),
                (
                    "two",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "test"]
                            )
                        ),
                        ref_values={"Environment": "test"},
                    ),
                    None,
                ),
                (
                    "three",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "prod"]
                            )
                        ),
                        ref_values={"Environment": "prod"},
                    ),
                    None,
                ),
            ],
        ),
        (
            "Valid FindInMap with a Ref to a parameter for top level key",
            {"Fn::FindInMap": ["environments", {"Ref": "RandomString"}, "foo"]},
            [],
        ),
        (
            "Valid FindInMap with a Ref to a parameter and a Ref to pseudo parameter",
            {
                "Fn::FindInMap": [
                    "accounts",
                    {"Ref": "AWS::AccountId"},
                    {"Ref": "RandomString"},
                ]
            },
            [],
        ),
        (
            "Valid FindInMap with a bad second key and default",
            {"Fn::FindInMap": ["foo", "first", "third", {"DefaultValue": "default"}]},
            [
                (
                    "default",
                    Context(path=Path(value_path=deque([4, "DefaultValue"]))),
                    None,
                )
            ],
        ),
        (
            "Valid FindInMap with a transform on first key",
            {"Fn::FindInMap": ["transformFirstKey", "first", "third"]},
            [],
        ),
        (
            "Valid FindInMap with a transform on second key",
            {"Fn::FindInMap": ["transformSecondKey", "first", "third"]},
            [],
        ),
        (
            "Valid FindInMap using a Sub",
            {
                "Fn::FindInMap": [
                    "environments",
                    "lion",
                    {"Fn::Sub": "${AWS::AccountId}Extra"},
                ]
            },
            [],
        ),
        (
            ("Valid FindInMap with a Sub with no parameters"),
            {"Fn::FindInMap": ["environments", "lion", {"Fn::Sub": "dev"}]},
            [
                (
                    "one",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "dev"]
                            )
                        )
                    ),
                    None,
                ),
            ],
        ),
        (
            ("Valid FindInMap with sub to a paremter"),
            {"Fn::FindInMap": ["environments", "lion", {"Fn::Sub": "${Environment}"}]},
            [
                (
                    "one",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "dev"]
                            )
                        ),
                        ref_values={"Environment": "dev"},
                    ),
                    None,
                ),
                (
                    "two",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "test"]
                            )
                        ),
                        ref_values={"Environment": "test"},
                    ),
                    None,
                ),
                (
                    "three",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "prod"]
                            )
                        ),
                        ref_values={"Environment": "prod"},
                    ),
                    None,
                ),
            ],
        ),
        (
            ("Valid FindInMap with sub list value to a paramter"),
            {
                "Fn::FindInMap": [
                    "environments",
                    "lion",
                    {"Fn::Sub": ["${Environment}", {}]},
                ]
            },
            [
                (
                    "one",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "dev"]
                            )
                        ),
                        ref_values={"Environment": "dev"},
                    ),
                    None,
                ),
                (
                    "two",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "test"]
                            )
                        ),
                        ref_values={"Environment": "test"},
                    ),
                    None,
                ),
                (
                    "three",
                    Context(
                        path=Path(
                            value_path=deque(
                                ["Mappings", "environments", "lion", "prod"]
                            )
                        ),
                        ref_values={"Environment": "prod"},
                    ),
                    None,
                ),
            ],
        ),
        (
            ("Valid FindInMap with an invalid sub"),
            {
                "Fn::FindInMap": [
                    "environments",
                    "lion",
                    {"Fn::Sub": {"A": "B", "C": "D"}},
                ]
            },
            [],
        ),
        (
            "Valid Sub with a resolvable values",
            {"Fn::Sub": ["${a}-${b}", {"a": "foo", "b": "bar"}]},
            [("foo-bar", Context(), None)],
        ),
        (
            "Valid Sub with empty parameters",
            {"Fn::Sub": ["foo", {}]},
            [("foo", Context(), None)],
        ),
        (
            "Valid Sub with a getatt and list",
            {"Fn::Sub": ["${MyResource.Arn}", {}]},
            [],
        ),
        (
            "Valid Sub with a getatt string",
            {"Fn::Sub": "${MyResource.Arn}"},
            [],
        ),
        (
            "Fn::Join uses previous values when doing resolution",
            {"Fn::Join": ["-", [{"Ref": "Environment"}, {"Ref": "Environment"}]]},
            [
                ("dev-dev", Context(ref_values={"Environment": "dev"}), None),
                ("test-test", Context(ref_values={"Environment": "test"}), None),
                ("prod-prod", Context(ref_values={"Environment": "prod"}), None),
            ],
        ),
        (
            "Fn::Join using a few values with a bad Ref",
            {"Fn::Join": ["-", [{"Ref": "Environment"}, {"Ref": "DNE"}]]},
            [],
        ),
    ],
)
def test_valid_functions(name, instance, response):
    context = Context(
        parameters={
            "RandomString": Parameter(
                {
                    "Type": "String",
                }
            ),
            "Environment": Parameter(
                {
                    "Type": "String",
                    "AllowedValues": ["dev", "test", "prod"],
                }
            ),
        },
        mappings=Mappings.create_from_dict(
            {
                "foo": {"first": {"second": "bar"}},
                "transformFirstKey": {"Fn::Transform": {"second": "bar"}},
                "transformSecondKey": {"first": {"Fn::Transform": "bar"}},
                "integers": {"1": {"2": "Value"}},
                "accounts": {
                    "123456789012": {"dev": "bar"},
                    "us-east-1": {"bar": "foo"},
                },
                "environments": {
                    "dev": {"foo": "one"},
                    "test": {"foo": "two"},
                    "prod": {"bar": "three"},
                    "lion": {
                        "dev": "one",
                        "test": "two",
                        "prod": "three",
                    },
                },
            }
        ),
        resources={
            "MyResource": Resource(
                {
                    "Type": "AWS::S3::Bucket",
                    "Properties": {
                        "BucketName": "XXX",
                    },
                }
            ),
        },
    )
    _resolve(name, instance, response, context=context)


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Invalid FindInMap with no mappings",
            {"Fn::FindInMap": [{"Ref": "MyParameter"}, "B", "C"]},
            [
                (
                    None,
                    Context(),
                    ValidationError(
                        ("{'Ref': 'MyParameter'} is not one of []"),
                        path=deque(["Fn::FindInMap", 0]),
                    ),
                )
            ],
        ),
        (
            "Invalid FindInMap with no mappings and default value",
            {"Fn::FindInMap": ["A", "B", "C", {"DefaultValue": "default"}]},
            [
                (
                    "default",
                    Context(path=Path(value_path=deque([4, "DefaultValue"]))),
                    None,
                )
            ],
        ),
    ],
)
def test_no_mapping(name, instance, response):
    _resolve(name, instance, response)


@pytest.mark.parametrize(
    "name,instance,response",
    [
        (
            "Valid FindInMap using transform key (maybe) with fn",
            {"Fn::FindInMap": [{"Ref": "AWS::Region"}, "B", "C"]},
            [],
        ),
        (
            "Valid FindInMap using transform key (maybe)",
            {"Fn::FindInMap": ["A", "B", "C"]},
            [],
        ),
    ],
)
def test_find_in_map_with_transform(name, instance, response):
    context = Context(
        mappings=Mappings.create_from_dict(
            {
                "foo": {"first": {"second": "bar"}},
                "Fn::Transform": "foobar",
            }
        ),
        resources={},
    )
    _resolve(name, instance, response, context=context)


def test_if():
    cfn = Template(
        None,
        {
            "Conditions": {
                "Condition": {"Fn::Equals": [{"Ref": "AWS::StackName"}, "test"]}
            },
            "Resources": {},
        },
    )
    context = create_context_for_template(cfn)
    validator = CfnTemplateValidator().evolve(
        context=context,
        cfn=cfn,
    )

    instance = {"Fn::If": ["Condition", "Foo", "Bar"]}

    results = list(validator.resolve_value(instance))

    assert len(results) == 2
    assert results[0][0] == "Foo"
    assert results[1][0] == "Bar"

    assert results[0][1].context.conditions.status == {"Condition": True}
    assert results[1][1].context.conditions.status == {"Condition": False}

    validator = validator.evolve(
        context=validator.context.evolve(
            conditions=validator.context.conditions.evolve({"Condition": False})
        )
    )

    results = list(validator.resolve_value(instance))
    assert len(results) == 1
    assert results[0][0] == "Bar"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import cfnlint.data.schemas.extensions.aws_elasticache_cachecluster
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class CacheClusterEngine(CfnLintJsonSchema):
    id = "E3695"
    shortdesc = "Validate Elasticache Cluster Engine and Engine Version"
    description = (
        "Validate the Elasticache cluster engine along with the engine version"
    )
    tags = ["resources"]
    source_url = "https://docs.aws.amazon.com/AmazonElastiCache/latest/dg/supported-engine-versions.html"

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ElastiCache::CacheCluster/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_elasticache_cachecluster,
                filename="engine_version.json",
            ),
            all_matches=True,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.iam.Permissions import Permissions


@pytest.fixture(scope="module")
def rule():
    rule = Permissions()
    yield rule


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid string",
            "s3:GetObject",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "Valid list",
            ["s3:GetObject"],
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "Invalid string",
            "s3:Foo",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [
                ValidationError(
                    "'foo' is not one of ['getobject', 'listaccesspoints']",
                    rule=Permissions(),
                )
            ],
        ),
        (
            "Invalid list",
            ["s3:Foo"],
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [
                ValidationError(
                    "'foo' is not one of ['getobject', 'listaccesspoints']",
                    rule=Permissions(),
                )
            ],
        ),
        (
            "Valid astrisk for permission",
            ["s3:*"],
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "Valid all astrisk",
            ["*"],
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "Valid string with ending astrisk",
            "s3:Get*",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "Valid string with starting astrisk",
            "s3:*Object",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "Invalid list",
            ["s3:Foo"],
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [
                ValidationError(
                    "'foo' is not one of ['getobject', 'listaccesspoints']",
                    rule=Permissions(),
                )
            ],
        ),
        (
            "Invalid string with ending astrisk",
            "s3:Foo*",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [
                ValidationError(
                    "'foo*' is not one of ['getobject', 'listaccesspoints']",
                    rule=Permissions(),
                )
            ],
        ),
        (
            "Invalid string with starting astrisk",
            "s3:*Foo",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [
                ValidationError(
                    "'*foo' is not one of ['getobject', 'listaccesspoints']",
                    rule=Permissions(),
                )
            ],
        ),
        (
            "Invalid service",
            "foo:Bar",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [ValidationError("'foo' is not one of ['iam', 's3']", rule=Permissions())],
        ),
        (
            "Empty string",
            "",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [
                ValidationError(
                    (
                        "'' is not a valid action. Must be "
                        "of the form service:action or '*'"
                    ),
                    rule=Permissions(),
                )
            ],
        ),
        (
            "A function",
            {"Ref": "MyParameter"},
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "asterisk in the middle",
            "iam:*Tags",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "multiple asterisks good",
            "iam:*Group*",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "multiple asterisks bad",
            "iam:*ec2*",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [
                ValidationError(
                    "'*ec2*' is not one of ['tagrole', 'creategroup', 'listgrouptags']",
                    rule=Permissions(),
                )
            ],
        ),
        (
            "question mark is bad",
            "iam:Tag?",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [
                ValidationError(
                    "'tag?' is not one of ['tagrole', 'creategroup', 'listgrouptags']",
                    rule=Permissions(),
                )
            ],
        ),
        (
            "question mark is good",
            "iam:TagRol?",
            {"cfn_path": deque(["Resources", "AWS::IAM::ManagedPolicy", "Properties"])},
            [],
        ),
        (
            "valid s3 bucket action",
            "s3:getobject",
            {"cfn_path": deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])},
            [],
        ),
        (
            "invalid s3 bucket action",
            "iam:tagrole",
            {"cfn_path": deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])},
            [ValidationError("'iam' is not one of ['s3']", rule=Permissions())],
        ),
        (
            "invalid s3 bucket action bucket short path",
            "iam:tagrole",
            {"cfn_path": deque(["Resources"])},
            [],
        ),
        (
            "invalid s3 bucket action",
            "iam:tagrole",
            {"cfn_path": deque(["Resources", "AWS::KMS::KeyPolicy", "Properties"])},
            [],
        ),
    ],
    indirect=["path"],
)
def test_permissions(name, instance, expected, rule, validator):

    rule._service_map = {
        "iam": {
            "Actions": {
                "tagrole": {"Resources": ["role"]},
                "creategroup": {"Resources": ["group"]},
                "listgrouptags": {"Resources": ["group"]},
            },
            "Resources": {
                "role": {
                    "ARNFormats": ["arn:${Partition}:iam::${Account}:role/.*"],
                    "ConditionKeys": [
                        "aws:ResourceTag/${TagKey}",
                        "iam:ResourceTag/${TagKey}",
                    ],
                },
                "group": {"ARNFormats": ["arn:${Partition}:iam::${Account}:group/.*"]},
            },
        },
        "s3": {
            "Actions": {
                "getobject": {"Resources": ["Object"]},
                "listaccesspoints": {},
            },
            "Resources": {
                "object": {"ARNFormats": ["arn:${Partition}:s3:::.*"]},
            },
        },
    }

    errors = list(rule.validate(validator, {}, instance, {}))

    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import regex as re

from cfnlint.helpers import REGEX_SUB_PARAMETERS, TRANSFORM_SAM, ensure_list
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules import CloudFormationLintRule


class SubUnneeded(CloudFormationLintRule):
    """Check if Sub is using a variable"""

    id = "W1020"
    shortdesc = "Sub isn't needed if it doesn't have a variable defined"
    description = "Checks sub strings to see if a variable is defined."
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub"]

    def validate(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        if TRANSFORM_SAM in ensure_list(validator.cfn.transform_pre.get("Transform")):
            return
        if validator.is_type(instance, "string"):
            variables = re.findall(REGEX_SUB_PARAMETERS, instance)
            path = []
        else:
            variables = re.findall(REGEX_SUB_PARAMETERS, instance[0])
            path = [0]
        if not variables:
            yield ValidationError(
                "'Fn::Sub' isn't needed because there are no variables", path=path
            )



"""
Helpers for loading resources, managing specs, constants, etc.

Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
import logging
import warnings
from copy import deepcopy
from typing import Any

import networkx
import regex as re

LOGGER = logging.getLogger("cfnlint.graph")


class EdgeSetting:
    color: str = "black"
    labal: str = ""

    def __init__(self, color: str, label: str) -> None:
        self.color = color
        self.label = label


class NodeSetting:
    color: str = "black"
    shape: str
    node_type: str

    def __init__(self, color: str, node_type: str, shape: str = "ellipse") -> None:
        self.color = color
        self.shape = shape
        self.node_type = node_type


class GraphSettings:
    ref: EdgeSetting
    getatt: EdgeSetting
    depends_on: EdgeSetting
    resource: NodeSetting
    parameter: NodeSetting
    output: NodeSetting

    def _pydot_string_convert(self, value: str | int) -> str | int:
        if not isinstance(value, str):
            return value
        if ":" in value:
            return f'"{value}"'

        return value

    def _pydot_list_convert(self, value: list[str | int]) -> str:
        return json.dumps(json.dumps(value))

    def subgraph_view(self, graph) -> networkx.MultiDiGraph:
        view = networkx.MultiDiGraph(name="template")
        resources: list[str] = [
            n for n, v in graph.nodes.items() if v["type"] in ["Resource"]
        ]

        # have to add quotes when outputing to dot
        for resource in resources:
            node = graph.nodes[resource]
            node["label"] = self._pydot_string_convert(node["label"])
            node["type"] = self._pydot_string_convert(node["type"])
            del node["resource_type"]
            view.add_node(resource, **node)

        for edge_1, edge_2, edge_data in graph.edges(data=True):
            if edge_1 in resources and edge_2 in resources:
                edge_data["source_paths"] = self._pydot_list_convert(
                    edge_data["source_paths"]
                )
                view.add_edge(
                    edge_1,
                    edge_2,
                    **edge_data,
                )

        view.graph.update(graph.graph)
        return view


class DefaultGraphSettings(GraphSettings):
    ref = EdgeSetting(color="black", label="Ref")
    getatt = EdgeSetting(color="black", label="GetAtt")
    depends_on = EdgeSetting(color="black", label="DependsOn")
    resource = NodeSetting(color="black", node_type="Resource")
    parameter = NodeSetting(color="black", node_type="Parameter", shape="box")
    output = NodeSetting(color="black", node_type="Output", shape="box")


class Graph:
    """Models a template as a directed graph of resources"""

    settings: GraphSettings
    __supported_types: list[str] = ["Resources", "Parameters", "Outputs"]

    def __init__(self, cfn):
        """Builds a graph where resources are nodes and edges are
        explicit (DependsOn) or implicit (Fn::GetAtt, Fn::Sub, Ref)
        relationships between resources
        """

        self.settings = DefaultGraphSettings()

        # Directed graph that allows self loops and parallel edges
        self.graph = networkx.MultiDiGraph(name="template")

        self._add_resources(cfn)
        self._add_parameters(cfn)
        self._add_outputs(cfn)
        self._add_refs(cfn)
        self._add_getatts(cfn)
        self._add_subs(cfn)

    def get_cycles(self, cfn):
        """Return all resource pairs that have a cycle in them"""
        result = []
        for starting_resource in cfn.template.get("Resources", {}):
            try:
                for edge in list(networkx.find_cycle(self.graph, starting_resource)):
                    if edge not in result:
                        result.append(edge)
            except networkx.NetworkXNoCycle:
                continue
        return result

    def _add_parameters(self, cfn: Any) -> None:
        # add all parameters in the template as nodes
        for parameter_id, parameter_values in cfn.template.get(
            "Parameters", {}
        ).items():
            if not isinstance(parameter_values, dict):
                continue
            type_val = parameter_values.get("Type", "")
            if not isinstance(type_val, str):
                continue
            graph_label = str.format(f'"{parameter_id}\\n<{type_val}>"')
            self._add_node(
                parameter_id, label=graph_label, settings=self.settings.parameter
            )

    def _add_outputs(self, cfn: Any) -> None:
        # add all outputs in the template as nodes
        outputs = cfn.template.get("Outputs", {})
        if not isinstance(outputs, dict):
            return
        for output_id in outputs.keys():
            graph_label = str.format(f'"{output_id}"')
            self._add_node(output_id, settings=self.settings.output, label=graph_label)

    def _add_resources(self, cfn: Any):
        # add all resources in the template as nodes
        for resourceId, resourceVals in cfn.template.get("Resources", {}).items():
            if not isinstance(resourceVals, dict):
                continue
            type_val = resourceVals.get("Type", "")
            if not isinstance(type_val, str):
                continue
            graph_label = str.format(f"{resourceId}\\n<{type_val}>")
            self._add_node(
                resourceId,
                settings=self.settings.resource,
                label=graph_label,
                resource_type=type_val,
            )
            target_ids = resourceVals.get("DependsOn", [])
            if isinstance(target_ids, (list, str)):
                if isinstance(target_ids, (str)):
                    target_ids = [target_ids]
                for target_id in target_ids:
                    if isinstance(target_id, str):
                        if self._is_resource(cfn, target_id):
                            self._add_edge(
                                resourceId,
                                target_id,
                                ["DependsOn"],
                                self.settings.depends_on,
                            )

    def _add_refs(self, cfn: Any) -> None:
        # add edges for "Ref" tags. { "Ref" : "logicalNameOfResource" }
        refs_paths = cfn.search_deep_keys("Ref")
        for ref_path in refs_paths:
            ref_type, source_id = ref_path[:2]
            source_path = ref_path[2:-2]
            target_id = ref_path[-1]
            if ref_type not in self.__supported_types:
                continue

            if ref_type in ["Parameters", "Outputs"]:
                source_id = f"{ref_type[:-1]}-{source_id}"

            if isinstance(target_id, (str, int)) and (
                self._is_resource(cfn, target_id)
            ):
                self._add_edge(source_id, target_id, source_path, self.settings.ref)

    def _add_getatts(self, cfn: Any) -> None:
        # add edges for "Fn::GetAtt" tags.
        # { "Fn::GetAtt" : [ "logicalNameOfResource", "attributeName" ] }
        # or { "!GetAtt" : "logicalNameOfResource.attributeName" }
        getatt_paths = cfn.search_deep_keys("Fn::GetAtt")
        for getatt_path in getatt_paths:
            ref_type, source_id = getatt_path[:2]
            source_path = getatt_path[2:-2]
            value = getatt_path[-1]
            if ref_type not in self.__supported_types:
                continue

            if ref_type in ["Parameters", "Outputs"]:
                source_id = f"{ref_type[:-1]}-{source_id}"

            if (
                isinstance(value, list)
                and len(value) == 2
                and (self._is_resource(cfn, value[0]))
            ):
                target_resource_id = value[0]
                self._add_edge(
                    source_id, target_resource_id, source_path, self.settings.getatt
                )

            if isinstance(value, (str, str)) and "." in value:
                target_resource_id = value.split(".")[0]
                if self._is_resource(cfn, target_resource_id):
                    self._add_edge(
                        source_id, target_resource_id, source_path, self.settings.getatt
                    )

    def _add_subs(self, cfn: Any) -> None:
        # add edges for "Fn::Sub" tags.
        # E.g. { "Fn::Sub": "arn:aws:ec2:${AWS::Region}:${AWS::AccountId}:vpc/${vpc}" }
        sub_objs = cfn.search_deep_keys("Fn::Sub")
        for sub_obj in sub_objs:
            sub_parameters = []
            sub_parameter_values = {}
            value = sub_obj[-1]
            source_path = sub_obj[2:-2]
            ref_type, source_id = sub_obj[:2]

            if ref_type not in self.__supported_types:
                continue

            if ref_type in ["Parameters", "Outputs"]:
                source_id = f"{ref_type[:-1]}-{source_id}"

            if isinstance(value, list):
                if not value:
                    continue
                if len(value) == 2:
                    sub_parameter_values = value[1]
                sub_parameters = self._find_parameter(value[0])
            elif isinstance(value, (str)):
                sub_parameters = self._find_parameter(value)

            for sub_parameter in sub_parameters:
                if sub_parameter not in sub_parameter_values:
                    if "." in sub_parameter:
                        target_id = sub_parameter.split(".")[0]
                        if self._is_resource(cfn, target_id):
                            self._add_edge(
                                source_id, target_id, source_path, self.settings.getatt
                            )
                    elif self._is_resource(cfn, sub_parameter):
                        self._add_edge(
                            source_id, sub_parameter, source_path, self.settings.ref
                        )

    def _add_node(self, node_id, settings, **attr):
        if settings.node_type in ["Parameter", "Output"]:
            node_id = f"{settings.node_type}-{node_id}"

        attr.setdefault("color", settings.color)
        attr.setdefault("shape", settings.shape)
        attr.setdefault("type", settings.node_type)
        self.graph.add_node(node_id, **attr)

    def _add_edge(self, source_id, target_id, source_path, settings):
        self.graph.add_edge(
            source_id,
            target_id,
            source_paths=source_path,
            label=settings.label,
            color=settings.color,
        )

    def _is_resource(self, cfn, identifier):
        """Check if the identifier is that of a Resource"""
        if not isinstance(identifier, str):
            return {}
        return cfn.template.get("Resources", {}).get(identifier, {})

    def _find_parameter(self, string):
        """Search string for tokenized fields"""
        regex = re.compile(r"\${([a-zA-Z0-9.]*)}")
        if isinstance(string, str):
            return regex.findall(string)

        return []

    # pylint: disable=import-outside-toplevel,unused-variable
    def to_dot(self, path):
        """Export the graph to a file with DOT format"""
        graph = deepcopy(self.graph)
        view = self.settings.subgraph_view(graph)
        try:
            networkx.drawing.nx_agraph.write_dot(view, path)
        except ImportError:
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=PendingDeprecationWarning)
                    warnings.simplefilter("ignore", category=DeprecationWarning)

                    networkx.drawing.nx_pydot.write_dot(view, path)
            except ImportError as e:
                raise e



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from unittest.mock import ANY, MagicMock

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.templates.LimitDescription import LimitDescription


@pytest.fixture(scope="module")
def rule():
    rule = LimitDescription()
    yield rule


@pytest.mark.parametrize(
    "name,instance,add_rule,child_return_values,expected",
    [
        (
            "Valid description",
            "Foo",
            True,
            [],
            [],
        ),
        (
            "Invalid type",
            {},
            True,
            None,
            [],
        ),
        (
            "Too long",
            "FooBar",
            True,
            None,
            [
                ValidationError(
                    ("expected maximum length: 3, found: 6"),
                )
            ],
        ),
        (
            "Child rule good",
            "Foo",
            True,
            [],
            [],
        ),
        (
            "Child rule bad",
            "Foo",
            True,
            [ValidationError("Bad")],
            [
                ValidationError(
                    ("Bad"),
                )
            ],
        ),
        (
            "Child rule None",
            "Foo",
            False,
            None,
            [],
        ),
    ],
)
def test_validate(
    name, instance, add_rule, child_return_values, expected, rule, validator
):
    child_rule = MagicMock()
    if child_return_values is not None:
        child_rule.maxLength.return_value = iter(child_return_values)

    if add_rule:
        rule.child_rules["I1003"] = child_rule

    errors = list(rule.maxLength(validator, 3, instance, {}))

    assert errors == expected, f"Test {name!r} got {errors!r}"

    if child_return_values is not None:
        assert child_rule.maxLength.mock_calls == [ANY]
    else:
        assert child_rule.maxLength.mock_calls == []



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from unittest import TestCase

from cfnlint.context import Context
from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.iam.IdentityPolicy import IdentityPolicy


class TestIdentityPolicies(TestCase):
    """Test IAM identity Policies"""

    def setUp(self):
        """Setup"""
        self.rule = IdentityPolicy()

    def test_object_basic(self):
        """Test Positive"""
        validator = CfnTemplateValidator()

        policy = {"Version": "2012-10-18"}

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(errs[0].message, "'Statement' is a required property")
        self.assertListEqual(list(errs[0].path), [])
        self.assertEqual(len(errs), 2, errs)
        self.assertEqual(
            errs[1].message, "'2012-10-18' is not one of ['2008-10-17', '2012-10-17']"
        )
        self.assertListEqual(list(errs[1].path), ["Version"])

    def test_object_multiple_effect(self):
        validator = CfnTemplateValidator()

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "NotAction": "*",
                    "Action": [
                        "cloudformation:*",
                    ],
                    "Resource": "*",
                }
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 2, errs)
        self.assertEqual(
            errs[0].message,
            ("Only one of ['Action', 'NotAction'] is a required property"),
        )
        self.assertEqual(
            errs[1].message,
            ("Only one of ['Action', 'NotAction'] is a required property"),
        )
        self.assertIn(
            ["Statement", 0, "NotAction"], [list(errs[0].path), list(errs[1].path)]
        )
        self.assertIn(
            ["Statement", 0, "Action"], [list(errs[0].path), list(errs[1].path)]
        )

    def test_object_statements(self):
        validator = CfnTemplateValidator({}).evolve(
            context=Context(functions=FUNCTIONS)
        )

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "NotAllow",
                    "Action": [
                        "cloudformation:Describe*",
                        "cloudformation:List*",
                        "cloudformation:Get*",
                    ],
                    "Resource": [
                        {
                            "Fn::Sub": "arn:${AWS::Partition}:iam::123456789012:role/object-role"
                        },
                        {
                            "Fn::Sub": "arn:aws:cloudformation:${AWS::Region}:aws:transform/Serverless-2016-10-31"
                        },
                        {
                            "NotValid": [
                                "arn:${AWS::Partition}:iam::123456789012:role/object-role"
                            ]
                        },
                        "arn:aws:medialive:*",
                    ],
                }
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 2, errs)
        self.assertEqual(errs[0].message, "'NotAllow' is not one of ['Allow', 'Deny']")
        self.assertListEqual(list(errs[0].path), ["Statement", 0, "Effect"])
        self.assertEqual(
            errs[1].message,
            "{'NotValid': ['arn:${AWS::Partition}:iam::123456789012:role/object-role']} is not of type 'string'",
        )
        self.assertListEqual(list(errs[1].path), ["Statement", 0, "Resource", 2])

    def test_string_statements(self):
        """Test Positive"""
        validator = CfnTemplateValidator()

        # ruff: noqa: E501
        policy = """
            {
                "Version": "2012-10-18",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Action": [
                            "cloudformation:Describe*",
                            "cloudformation:List*",
                            "cloudformation:Get*"
                        ],
                        "Resource": [
                            "*",
                            {"Fn::Sub": ["arn:${AWS::Partition}:iam::123456789012/role/string-role"]}
                        ]
                    }
                ]
            }
        """

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 2, errs)
        self.assertEqual(
            errs[0].message,
            "{'Fn::Sub': ['arn:${AWS::Partition}:iam::123456789012/role/string-role']} is not of type 'string'",
        )
        self.assertListEqual(list(errs[0].path), ["Statement", 0, "Resource", 1])
        self.assertEqual(
            errs[1].message, "'2012-10-18' is not one of ['2008-10-17', '2012-10-17']"
        )
        self.assertListEqual(list(errs[1].path), ["Version"])

    def test_string_statements_with_condition(self):
        validator = CfnTemplateValidator()

        policy = """
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Action": "*",
                        "Resource": "*",
                        "Condition": {
                            "iam:PassedToService": "cloudformation.amazonaws.com",
                            "StringEquals": {"aws:PrincipalTag/job-category": "iamuser-admin"},
                            "StringLike": {"s3:prefix": ["", "home/", "home/${aws:username}/"]},
                            "ArnLike": {"aws:SourceArn": "arn:aws:cloudtrail:*:111122223333:trail/*"},
                            "NumericLessThanEquals": {"s3:max-keys": "10"},
                            "DateGreaterThan": {"aws:TokenIssueTime": "2020-01-01T00:00:01Z"},
                            "Bool": { "aws:SecureTransport": "false"},
                            "BinaryEquals": { "key" : "QmluYXJ5VmFsdWVJbkJhc2U2NA=="},
                            "IpAddress": {"aws:SourceIp": "203.0.113.0/24"},
                            "ArnEquals": {"aws:SourceArn": "arn:aws:sns:REGION:123456789012:TOPIC-ID"},
                            "StringLikeIfExists": { "ec2:InstanceType": [ "t1.*", "t2.*" ]},
                            "Null":{"aws:TokenIssueTime":"true"},
                            "ForAllValues:StringEquals":{"aws:PrincipalTag/job-category":["iamuser-admin","iamuser-read-only"]},
                            "ForAnyValue:StringEquals":{"aws:PrincipalTag/job-category":"iamuser-admin"},
                            "ForAllValues:StringLike": {"aws:TagKeys": ["key1*"]}
                        }
                    }
                ]
            }
        """

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 1, errs)
        self.assertTrue(
            errs[0].message.startswith("'iam:PassedToService' does not match")
        )
        self.assertListEqual(
            list(errs[0].path), ["Statement", 0, "Condition", "iam:PassedToService"]
        )

    def test_duplicate_sid(self):
        validator = CfnTemplateValidator()

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "All",
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": "*",
                },
                {
                    "Sid": "All",
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": "*",
                },
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 1, errs)
        self.assertEqual(errs[0].message, "array items are not unique for keys ['Sid']")
        self.assertListEqual(list(errs[0].path), ["Statement"])

    def test_pattern_sid(self):
        validator = CfnTemplateValidator()

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "A ",
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": "*",
                },
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertListEqual(
            errs,
            [
                ValidationError(
                    message="'A ' does not match '^[A-Za-z0-9]+$'",
                    validator="pattern",
                    path=deque(["Statement", 0, "Sid"]),
                    schema_path=deque(
                        [
                            "properties",
                            "Statement",
                            "items",
                            "properties",
                            "Sid",
                            "pattern",
                        ]
                    ),
                    rule=IdentityPolicy(),
                )
            ],
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class BackupPlanLifecycleRule(CfnLintKeyword):
    """
    Check Backup Plan rules with lifecycle has minimum
    period between cold and delete
    """

    id = "E3504"
    shortdesc = "Check minimum 90 period is met between BackupPlan cold and delete"
    description = (
        "Check that Backup plans with lifecycle rules have >= 90 days between cold and"
        " delete"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-backup-backupplan-lifecycleresourcetype.html"
    tags = ["properties", "backup", "plan", "lifecycle"]

    def __init__(self) -> None:
        super().__init__(
            [
                "Resources/AWS::Backup::BackupPlan/Properties/BackupPlan/BackupPlanRule/*/Lifecycle",
                "Resources/AWS::Backup::BackupPlan/Properties/BackupPlan/BackupPlanRule/*/CopyActions/*/Lifecycle",
            ]
        )

    def validate(
        self, validator: Validator, uI: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        delete_after_days = instance.get("DeleteAfterDays")
        move_to_cold_storage_after_days = instance.get("MoveToColdStorageAfterDays")

        if not validator.is_type(delete_after_days, "integer"):
            return
        if not validator.is_type(move_to_cold_storage_after_days, "integer"):
            return
        if delete_after_days - move_to_cold_storage_after_days < 90:
            yield ValidationError(
                (
                    f"DeleteAfterDays {delete_after_days!r} must be at least "
                    "90 days after MoveToColdStorageAfterDays "
                    f"{move_to_cold_storage_after_days}"
                ),
                path=deque(["DeleteAfterDays"]),
                rule=self,
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules import CfnLintKeyword
from cfnlint.rules.functions.GetAtt import GetAtt


@pytest.fixture(scope="module")
def rule():
    rule = GetAtt()
    yield rule


_template = {
    "Resources": {
        "MyBucket": {"Type": "AWS::S3::Bucket"},
        "MyCodePipeline": {"Type": "AWS::CodePipeline::Pipeline"},
        "DocDBCluster": {"Type": "AWS::DocDB::DBCluster"},
    },
    "Parameters": {
        "MyResourceParameter": {"Type": "String", "Default": "MyBucket"},
        "MyAttributeParameter": {"Type": "String", "AllowedValues": ["Arn"]},
    },
}

_template_with_transform = _template.copy()
_template_with_transform["Transform"] = "AWS::LanguageExtensions"


class _Pass(CfnLintKeyword):
    id = "AAAAA"

    def __init__(self) -> None:
        super().__init__(keywords=["*"])

    def validate(self, validator, s, instance, schema):
        return
        yield


class _Fail(CfnLintKeyword):
    id = "BBBBB"

    def __init__(self) -> None:
        super().__init__(keywords=["*"])

    def validate(self, validator, s, instance, schema):
        yield ValidationError("Fail")


@pytest.mark.parametrize(
    "name,instance,schema,template,child_rules,expected",
    [
        (
            "Valid GetAtt with a good attribute",
            {"Fn::GetAtt": ["MyBucket", "Arn"]},
            {"type": "string"},
            _template,
            {},
            [],
        ),
        (
            "Invalid GetAtt with bad attribute",
            {"Fn::GetAtt": ["MyBucket", "foo"]},
            {"type": "string"},
            _template,
            {},
            [
                ValidationError(
                    (
                        "'foo' is not one of ['Arn', 'DomainName', "
                        "'DualStackDomainName', 'RegionalDomainName', "
                        "'MetadataTableConfiguration.S3TablesDestination.TableNamespace',"
                        " 'MetadataTableConfiguration.S3TablesDestination.TableArn',"
                        " 'WebsiteURL'] "
                        "in ['us-east-1']"
                    ),
                    path=deque(["Fn::GetAtt", 1]),
                    schema_path=deque([]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Invalid GetAtt with bad resource name",
            {"Fn::GetAtt": ["Foo", "bar"]},
            {"type": "string"},
            _template,
            {},
            [
                ValidationError(
                    (
                        "'Foo' is not one of ['MyBucket', "
                        "'MyCodePipeline', 'DocDBCluster']"
                    ),
                    path=deque(["Fn::GetAtt", 0]),
                    schema_path=deque(["enum"]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Invalid GetAtt with a bad type",
            {"Fn::GetAtt": {"foo": "bar"}},
            {"type": "string"},
            _template,
            {},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'string', 'array'",
                    path=deque(["Fn::GetAtt"]),
                    schema_path=deque(["cfnContext", "schema", "type"]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Invalid GetAtt with a bad response type",
            {"Fn::GetAtt": "MyBucket.Arn"},
            {"type": "array"},
            _template,
            {},
            [
                ValidationError(
                    "{'Fn::GetAtt': 'MyBucket.Arn'} is not of type 'array'",
                    path=deque(["Fn::GetAtt"]),
                    schema_path=deque(["type"]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Invalid GetAtt with a bad response type and multiple types",
            {"Fn::GetAtt": "MyBucket.Arn"},
            {"type": ["array", "object"]},
            _template,
            {},
            [
                ValidationError(
                    "{'Fn::GetAtt': 'MyBucket.Arn'} is not of type 'array', 'object'",
                    path=deque(["Fn::GetAtt"]),
                    schema_path=deque(["type"]),
                    validator="fn_getatt",
                ),
            ],
        ),
        (
            "Valid GetAtt with integer to string",
            {"Fn::GetAtt": "MyCodePipeline.Version"},
            {"type": ["integer"]},
            _template,
            {},
            [],
        ),
        (
            "Valid GetAtt with exception type",
            {"Fn::GetAtt": "DocDBCluster.Port"},
            {"type": ["string"]},
            _template,
            {},
            [],
        ),
        # (
        #    "Invalid GetAtt with integer to string",
        #    {"Fn::GetAtt": "MyCodePipeline.Version"},
        #    {"type": ["integer"]},
        #    {
        #        "strict_types": True,
        #    },
        #    {},
        #    [
        #        ValidationError(
        #            (
        #                "{'Fn::GetAtt': 'MyCodePipeline.Version'} "
        #                "is not of type 'integer'"
        #            ),
        #            path=deque(["Fn::GetAtt"]),
        #            schema_path=deque(["type"]),
        #            validator="fn_getatt",
        #        )
        #    ],
        # ),
        (
            "Valid GetAtt with one good response type",
            {"Fn::GetAtt": "MyBucket.Arn"},
            {"type": ["array", "string"]},
            _template,
            {},
            [],
        ),
        (
            "Valid Ref in GetAtt for resource",
            {"Fn::GetAtt": [{"Ref": "MyResourceParameter"}, "Arn"]},
            {"type": "string"},
            _template_with_transform,
            {},
            [],
        ),
        (
            "Valid Ref in GetAtt for attribute",
            {"Fn::GetAtt": ["MyBucket", {"Fn::Sub": "${MyAttributeParameter}"}]},
            {"type": "string"},
            _template_with_transform,
            {},
            [],
        ),
        (
            "Invalid Ref in GetAtt for attribute",
            {"Fn::GetAtt": ["MyBucket", {"Ref": "MyResourceParameter"}]},
            {"type": "string"},
            _template_with_transform,
            {},
            [
                ValidationError(
                    (
                        "'MyBucket' is not one of ['Arn', 'DomainName', "
                        "'DualStackDomainName', 'RegionalDomainName', "
                        "'MetadataTableConfiguration.S3TablesDestination.TableNamespace',"
                        " 'MetadataTableConfiguration.S3TablesDestination.TableArn',"
                        " 'WebsiteURL'] in ['us-east-1'] when "
                        "{'Ref': 'MyResourceParameter'} is resolved"
                    ),
                    path=deque(["Fn::GetAtt", 1]),
                    validator="fn_getatt",
                )
            ],
        ),
        (
            "Invalid Ref in GetAtt for attribute",
            {"Fn::GetAtt": [{"Ref": "MyAttributeParameter"}, "Arn"]},
            {"type": "string"},
            _template_with_transform,
            {},
            [
                ValidationError(
                    (
                        "'Arn' is not one of ['MyBucket', "
                        "'MyCodePipeline', 'DocDBCluster'] when "
                        "{'Ref': 'MyAttributeParameter'} is resolved"
                    ),
                    path=deque(["Fn::GetAtt", 0]),
                    schema_path=deque(["enum"]),
                    validator="fn_getatt",
                )
            ],
        ),
        (
            "Valid GetAtt with child rules",
            {"Fn::GetAtt": ["MyBucket", "Arn"]},
            {"type": "string"},
            _template,
            {
                "AAAAA": _Pass(),
                "BBBBB": _Fail(),
                "CCCCC": None,
            },
            [ValidationError("Fail")],
        ),
    ],
    indirect=["template"],
)
def test_validate(name, instance, schema, child_rules, expected, validator, rule):
    rule.child_rules = child_rules
    errs = list(rule.fn_getatt(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_rds_dbinstance
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class RetentionPeriodOnResourceTypesWithAutoExpiringContent(CfnLintJsonSchema):
    """Check Base Resource Configuration"""

    id = "I3013"
    shortdesc = (
        "Check resources with auto expiring content have explicit retention period"
    )
    description = (
        "The behaviour for data retention is different across AWS Services.If no"
        " retention period is specified the default for some services is to delete the"
        " data after a period of time.This check requires you to explicitly set the"
        " retention period for those resources to avoid unexpected data losses"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["resources", "retentionperiod"]

    def __init__(self) -> None:
        super().__init__(
            [
                "Resources/AWS::RDS::DBInstance/Properties",
                "Resources/AWS::Kinesis::Stream/Properties",
                "Resources/AWS::SQS::Queue/Properties",
                "Resources/AWS::DocDB::DBCluster/Properties",
                "Resources/AWS::Synthetics::Canary/Properties",
                "Resources/AWS::Redshift::Cluster/Properties",
                "Resources/AWS::RDS::DBInstance/Properties",
                "Resources/AWS::RDS::DBCluster/Properties",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_rds_dbinstance,
                filename="backupretentionperiod.json",
            ),
        )
        self._rds_schema = self._schema
        self._properties: dict[str, list[dict[str, str]]] = {
            "AWS::Kinesis::Stream": [
                {
                    "Attribute": "RetentionPeriodHours",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-kinesis-stream.html#cfn-kinesis-stream-retentionperiodhours",
                }
            ],
            "AWS::SQS::Queue": [
                {
                    "Attribute": "MessageRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-sqs-queues.html#aws-sqs-queue-msgretentionperiod",
                }
            ],
            "AWS::DocDB::DBCluster": [
                {
                    "Attribute": "BackupRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-docdb-dbcluster.html#cfn-docdb-dbcluster-backupretentionperiod",
                }
            ],
            "AWS::Synthetics::Canary": [
                {
                    "Attribute": "SuccessRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-synthetics-canary.html#cfn-synthetics-canary-successretentionperiod",
                },
                {
                    "Attribute": "FailureRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-synthetics-canary.html#cfn-synthetics-canary-failureretentionperiod",
                },
            ],
            "AWS::Redshift::Cluster": [
                {
                    "Attribute": "AutomatedSnapshotRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-redshift-cluster.html#cfn-redshift-cluster-automatedsnapshotretentionperiod",
                }
            ],
            "AWS::RDS::DBInstance": [
                {
                    "Attribute": "BackupRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-rds-database-instance.html#cfn-rds-dbinstance-backupretentionperiod",
                }
            ],
            "AWS::RDS::DBCluster": [
                {
                    "Attribute": "BackupRetentionPeriod",
                    "SourceUrl": "http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-rds-dbcluster.html#cfn-rds-dbcluster-backuprententionperiod",
                }
            ],
        }

    def message(self, instance: Any, err: ValidationError) -> str:
        return (
            f"{err.message} (The default retention period will delete "
            "the data after a pre-defined time. Set an explicit "
            "values to avoid data loss on resource)"
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if len(validator.context.path.cfn_path) < 1:
            return

        resource_type = validator.context.path.cfn_path[1]

        if resource_type != "AWS::RDS::DBInstance":
            required = []
            for properties in self._properties.get(resource_type, []):
                property = properties.get("Attribute")
                if property:
                    required.append(property)

            self._schema = {"required": required}
        else:
            self._schema = self._rds_schema

        yield from super().validate(validator, keywords, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from cfnlint.rules.formats.FormatKeyword import FormatKeyword


class IamRoleArn(FormatKeyword):
    id = "E1156"
    shortdesc = "Validate IAM role ARN format"
    description = "Validate IAM role ARN validation for ref/gett and string values"
    tags = []
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/format_keyword.md#AWS::IAM::Role.Arn"

    def __init__(self):
        super().__init__(
            format="AWS::IAM::Role.Arn",
            pattern=r"^arn:(aws|aws-cn|aws-iso|aws-iso-[a-z]{1}|aws-us-gov):iam::\d{12}:role/.*$",
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Iterable

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.codepipeline.PipelineFirstStageHasSource import (
    PipelineFirstStageHasSource,
)


@pytest.fixture(scope="module")
def rule():
    rule = PipelineFirstStageHasSource()
    yield rule


_standard_path = deque(["Resources", "TestPipeline", "Properties", "Stages"])


def _append_queues(queue1: Iterable, queue2: Iterable) -> deque:
    new_queue: deque = deque()
    new_queue.extend(queue1)
    new_queue.extend(queue2)
    return new_queue


@pytest.mark.parametrize(
    "instance,path,expected",
    [
        (
            [],
            {
                "path": _append_queues(
                    _standard_path, [0, "Actions", 0, "ActionTypeId", "Category"]
                ),
            },
            [],
        ),
        (
            "Source",
            {
                "path": _standard_path,
            },
            [],
        ),
        (
            "Source",
            {
                "path": _append_queues(
                    _standard_path, [0, "Actions", 0, "ActionTypeId", "Category"]
                ),
            },
            [],
        ),
        (
            "Build",
            {
                "path": _append_queues(
                    _standard_path, [1, "Actions", 0, "ActionTypeId", "Category"]
                ),
            },
            [],
        ),
        (
            "Source",
            {
                "path": _append_queues(
                    _standard_path,
                    ["Fn::If", 2, 0, "Actions", 0, "ActionTypeId", "Category"],
                ),
            },
            [],
        ),
        (
            "Build",
            {
                "path": _append_queues(
                    _standard_path,
                    ["Fn::If", 1, 1, "Actions", 0, "ActionTypeId", "Category"],
                ),
            },
            [],
        ),
        (
            "Build",
            {
                "path": _append_queues(
                    _standard_path, [0, "Actions", 0, "ActionTypeId", "Category"]
                ),
            },
            [
                ValidationError(
                    "'Build' is not one of ['Source']",
                    rule=PipelineFirstStageHasSource(),
                    validator="enum",
                )
            ],
        ),
        (
            "Build",
            {
                "path": _append_queues(
                    _standard_path,
                    ["Fn::If", 2, 0, "Actions", 0, "ActionTypeId", "Category"],
                ),
            },
            [
                ValidationError(
                    "'Build' is not one of ['Source']",
                    rule=PipelineFirstStageHasSource(),
                    validator="enum",
                )
            ],
        ),
        (
            "Source",
            {
                "path": _append_queues(
                    _standard_path, [1, "Actions", 0, "ActionTypeId", "Category"]
                ),
            },
            [
                ValidationError(
                    (
                        "'Source' is not one of ['Build', 'Approval', "
                        "'Deploy', 'Test', 'Invoke', 'Compute']"
                    ),
                    rule=PipelineFirstStageHasSource(),
                    validator="enum",
                )
            ],
        ),
        (
            "Source",
            {
                "path": _append_queues(
                    _standard_path,
                    ["Fn::If", 1, 1, "Actions", 0, "ActionTypeId", "Category"],
                ),
            },
            [
                ValidationError(
                    (
                        "'Source' is not one of ['Build', 'Approval', "
                        "'Deploy', 'Test', 'Invoke', 'Compute']"
                    ),
                    rule=PipelineFirstStageHasSource(),
                    validator="enum",
                )
            ],
        ),
    ],
    indirect=["path"],
)
def test_validate(instance, path, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.Condition import Condition
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = Condition()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {
            "Conditions": {
                "MyCondition": {"Fn::Equals": [{"Ref": "MyParam"}, "Foo"]},
            },
        },
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Condition is valid",
            "MyCondition",
            [],
        ),
        (
            "Valid because of wrong type",
            {},
            [],
        ),
        (
            "Invalid condition",
            "IsProduction",
            [
                ValidationError(
                    "'IsProduction' is not one of ['MyCondition']",
                    validator="enum",
                    schema_path=deque(["enum"]),
                    rule=Condition(),
                    path=deque([]),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, context, cfn):
    validator = CfnTemplateValidator(context=context, cfn=cfn)
    errs = list(rule.validate(validator, {}, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.functions.Cidr import Cidr
from cfnlint.rules.functions.ImportValue import ImportValue
from cfnlint.rules.functions.Join import Join
from cfnlint.rules.functions.Ref import Ref
from cfnlint.rules.functions.RefResolved import RefResolved
from cfnlint.rules.outputs.ImportValue import ImportValue as OutputsImportValue
from cfnlint.rules.outputs.Value import Value  # pylint: disable=E0401


@pytest.fixture
def template():
    return {
        "Parameters": {
            "additionalVpcCidr": {
                "Type": "String",
                "Default": "",
            },
            "badAdditionalVpcCidr": {
                "Type": "String",
                "Default": "1",
            },
        },
        "Conditions": {
            "isAdditionalVpc": {
                "Fn::Not": [
                    {
                        "Fn::Equals": [
                            {"Ref": "additionalVpcCidr"},
                            "",
                        ]
                    }
                ]
            },
            "isBadAdditionalVpc": {
                "Fn::Not": [
                    {
                        "Fn::Equals": [
                            {"Ref": "badAdditionalVpcCidr"},
                            "",
                        ]
                    }
                ]
            },
        },
    }


@pytest.fixture
def path():
    return Path(
        path=deque(["Outputs", "Test"]),
        value_path=deque(),
        cfn_path=deque(["Outputs", "*"]),
    )


@pytest.fixture
def validator(cfn, context):

    ref = Ref()
    ref.child_rules["W1030"] = RefResolved()

    importvalue = ImportValue()
    importvalue.child_rules["W6001"] = OutputsImportValue()

    yield CfnTemplateValidator(schema={}).extend(
        validators={
            "fn_join": Join().fn_join,
            "ref": ref.ref,
            "fn_cidr": Cidr().fn_cidr,
            "fn_importvalue": importvalue.fn_importvalue,
        }
    )(
        schema={},
        cfn=cfn,
        context=context,
    )


@pytest.mark.parametrize(
    "input,expected",
    [
        (
            {
                "Value": "foo",
            },
            [],
        ),
        (
            {
                "Value": 1.0,
            },
            [],
        ),
        (
            {
                "Value": 1,
            },
            [],
        ),
        (
            {"Value": True},
            [],
        ),
        (
            {"Value": [{}]},
            [
                ValidationError(
                    "[{}] is not of type 'string'",
                    validator="type",
                    schema_path=deque(["type"]),
                    path=deque(["Value"]),
                    rule=Value(),
                )
            ],
        ),
        (
            {"Value": {"foo": "bar"}},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'string'",
                    validator="type",
                    schema_path=deque(["type"]),
                    path=deque(["Value"]),
                    rule=Value(),
                )
            ],
        ),
        (
            {
                "Value": {"Fn::ImportValue": "test-stack-value"},
            },
            [
                ValidationError(
                    (
                        "The output value {'Fn::ImportValue': 'test-stack-value'} "
                        "is an import from another output"
                    ),
                    validator="fn_importvalue",
                    schema_path=deque(["fn_importvalue"]),
                    path=deque(["Value", "Fn::ImportValue"]),
                    rule=OutputsImportValue(),
                )
            ],
        ),
        (
            {
                "Condition": "isAdditionalVpc",
                "Value": {
                    "Fn::Join": [
                        ",",
                        {"Fn::Cidr": [{"Ref": "additionalVpcCidr"}, 3, 8]},
                    ]
                },
            },
            [],
        ),
        (
            {
                "Condition": "isBadAdditionalVpc",
                "Value": {
                    "Fn::Join": [
                        ",",
                        {"Fn::Cidr": [{"Ref": "badAdditionalVpcCidr"}, 3, 8]},
                    ]
                },
            },
            [
                ValidationError(
                    (
                        "{'Ref': 'badAdditionalVpcCidr'} does not match "
                        "'^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\\\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])(\\\\/([0-9]|[1-2][0-9]|3[0-2]))$'"
                        " when 'Ref' is resolved"
                    ),
                    validator="ref",
                    schema_path=deque(
                        [
                            "fn_join",
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            1,
                            "cfnContext",
                            "schema",
                            "fn_cidr",
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            0,
                            "cfnContext",
                            "schema",
                            "ref",
                            "pattern",
                        ]
                    ),
                    path=deque(["Value", "Fn::Join", 1, "Fn::Cidr", 0, "Ref"]),
                    rule=RefResolved(),
                )
            ],
        ),
    ],
)
def test_output_value(input, expected, validator):
    rule = Value()
    results = list(rule.validate(validator, {}, input, {}))
    assert results == expected, f"Expected {expected} results, got {results}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules.resources.iam.Policy import Policy


class ResourceEcrPolicy(Policy):
    """Check ECR repository policies"""

    id = "E3513"
    shortdesc = "Validate ECR repository policy"
    description = (
        "Private ECR repositories have a policy. This rule validates those policies."
    )
    source_url = "https://docs.aws.amazon.com/AmazonECR/latest/userguide/repository-policies.html"
    tags = ["resources", "iam", "ecr"]

    def __init__(self):
        super().__init__(
            [
                "Resources/AWS::ECR::Repository/Properties/RepositoryPolicyText",
            ],
            "resource",
            "policy_resource_ecr.json",
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from ipaddress import ip_network

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ectwo.VpcSubnetOverlap import VpcSubnetOverlap


@pytest.fixture(scope="module")
def rule():
    rule = VpcSubnetOverlap()
    yield rule


@pytest.fixture
def template():
    return {
        "Parameters": {"MyCidr": {"Type": "String"}},
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
            "IsUsWest2": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]},
            "IsUs": {
                "Fn::Or": [
                    {"Condition": "IsUsEast1"},
                    {"Condition": "IsUsWest2"},
                ]
            },
            "IsNotUs": {"Fn::Not": [{"Condition": "IsUs"}]},
        },
        "Resources": {},
    }


@pytest.mark.parametrize(
    "name,instance,starting_subnets,expected",
    [
        (
            "Valid with no overlap on ipv4",
            {"VpcId": {"Ref": "Vpc"}, "CidrBlock": "10.0.1.0/24"},
            {"Vpc": [(ip_network("10.0.0.0/24"), {})]},
            [],
        ),
        (
            "Valid with a bad cidr",
            {"VpcId": {"Ref": "Vpc"}, "CidrBlock": "10.0.0/24"},
            {"Vpc": [(ip_network("10.0.0.0/24"), {})]},
            [],
        ),
        (
            "Valid with no overlap on ipv6",
            {"VpcId": {"Fn::GetAtt": "Vpc.VpcId"}, "Ipv6CidrBlock": "2001:db8::/32"},
            {"Vpc": [(ip_network("2002:db8::/32"), {})]},
            [],
        ),
        (
            "Valid with a function",
            {"VpcId": {"Ref": "Vpc"}, "CidrBlock": {"Ref": "MyCidr"}},
            {"Vpc": [(ip_network("10.0.0.0/24"), {})]},
            [],
        ),
        (
            "Valid with no previous setting",
            {"VpcId": "vpc-123456", "CidrBlock": "10.0.0.0/24"},
            {},
            [],
        ),
        (
            "Valid with an invalid function",
            {"VpcId": {"DNE": "Vpc"}, "CidrBlock": "10.0.0/24"},
            {"Vpc": [(ip_network("10.0.0.0/24"), {})]},
            [],
        ),
        (
            "Valid with vpc using another function",
            {"VpcId": {"Fn::Join": ["vpc-1"]}, "CidrBlock": "10.0.1.0/24"},
            {"Vpc": [(ip_network("10.0.0.0/24"), {})]},
            [],
        ),
        (
            "Invalid with a overlap on ipv4",
            {"VpcId": {"Ref": "Vpc"}, "CidrBlock": "10.0.0.0/24"},
            {"Vpc": [(ip_network("10.0.0.0/22"), {})]},
            [
                ValidationError(
                    "'10.0.0.0/24' overlaps with '10.0.0.0/22'",
                    rule=VpcSubnetOverlap(),
                    path=deque(["CidrBlock"]),
                )
            ],
        ),
        (
            "Invalid with a overlap on ipv6",
            {"VpcId": {"Ref": "Vpc"}, "Ipv6CidrBlock": "fc00::/32"},
            {"Vpc": [(ip_network("fc00::/16"), {})]},
            [
                ValidationError(
                    "'fc00::/32' overlaps with 'fc00::/16'",
                    rule=VpcSubnetOverlap(),
                    path=deque(["Ipv6CidrBlock"]),
                )
            ],
        ),
        (
            "Valid with ipv4 and ipv6",
            {"VpcId": {"Ref": "Vpc"}, "Ipv6CidrBlock": "fc01::/32"},
            {
                "Vpc": [
                    (ip_network("fc00::/32"), {}),
                    (ip_network("10.0.0.0/24"), {}),
                ]
            },
            [],
        ),
        (
            "Valid with conflicting conditions",
            {
                "VpcId": {"Ref": "Vpc"},
                "CidrBlock": {
                    "Fn::If": ["IsNotUs", "10.0.0.0/24", {"Ref": "AWS::NoValue"}]
                },
            },
            {
                "Vpc": [
                    (
                        ip_network("10.0.0.0/24"),
                        {
                            "IsUsEast1": True,
                            "IsUsWest2": False,
                            "IsUs": True,
                            "IsNotUs": False,
                        },
                    ),
                ]
            },
            [],
        ),
        (
            "Invalid with conflicting conditions",
            {
                "VpcId": {"Ref": "Vpc"},
                "CidrBlock": {"Fn::If": ["IsNotUs", "10.0.1.0/24", "10.0.0.0/32"]},
            },
            {
                "Vpc": [
                    (
                        ip_network("10.0.0.0/24"),
                        {
                            "IsUsEast1": True,
                            "IsUsWest2": False,
                            "IsUs": True,
                            "IsNotUs": False,
                        },
                    ),
                ]
            },
            [
                ValidationError(
                    "'10.0.0.0/32' overlaps with '10.0.0.0/24'",
                    rule=VpcSubnetOverlap(),
                    path=deque(["CidrBlock", "Fn::If", 2]),
                )
            ],
        ),
    ],
    indirect=[],
)
def test_validate(name, instance, starting_subnets, expected, rule, validator):

    rule._subnets = starting_subnets
    errs = list(rule.validate(validator, "", instance, {}))

    assert (
        errs == expected
    ), f"Expected test {name!r} to have {expected!r} but got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
from test.testlib.testcase import BaseTestCase
from unittest.mock import call, mock_open, patch

import cfnlint.maintenance
from cfnlint.config import _DEFAULT_RULESDIR
from cfnlint.rules import Rules

LOGGER = logging.getLogger("cfnlint.maintenance")
LOGGER.addHandler(logging.NullHandler())


class TestUpdateDocumentation(BaseTestCase):
    """Used for Testing Updating the Documentation"""

    TEST_TEXT = """
Regular Text
## Rules
"""

    def test_update_docs(self):
        collection = Rules.create_from_directory(_DEFAULT_RULESDIR + "/errors")
        builtin_module_name = "builtins"

        mo = mock_open(read_data=self.TEST_TEXT)
        mo.return_value.__iter__ = lambda self: self
        mo.return_value.__iter__ = lambda self: iter(self.readline, "")
        with patch("{}.open".format(builtin_module_name), mo) as mock_builtin_open:
            cfnlint.maintenance.update_documentation(collection)

            expected_calls = [
                call("\n"),
                call("Regular Text\n"),
                call("## Rules\n"),
                call(
                    "(_This documentation is generated by running `cfn-lint"
                    " --update-documentation`, do not alter this manually_)\n\n"
                ),
                call(
                    "The following **{}** rules are applied by this linter:\n\n".format(
                        len(collection) + 3
                    )
                ),
                call(
                    "| Rule ID  | Title | Description | Config<br />(Name:Type:Default)"
                    " | Source | Tags |\n"
                ),
                call(
                    "| -------- | ----- | ----------- | ---------- | ------ | ---- |\n"
                ),
                call(
                    '| [E0000<a name="E0000"></a>]'
                    "(../src/cfnlint/rules/errors/parse.py) |"
                    " Parsing error found when parsing the template | Checks for"
                    " JSON/YAML formatting errors in your template |  |"
                    " [Source](https://github.com/aws-cloudformation/cfn-lint) |"
                    " `base` |\n"
                ),
                call(
                    '| [E0001<a name="E0001"></a>]'
                    "(../src/cfnlint/rules/errors/transform.py) |"
                    " Error found when transforming the template | Errors found when"
                    " performing transformation on the template |  |"
                    " [Source](https://github.com/aws-cloudformation/cfn-lint) |"
                    " `base`,`transform` |\n"
                ),
                call(
                    '| [E0002<a name="E0002"></a>]'
                    "(../src/cfnlint/rules/errors/rule.py) |"
                    " Error processing rule on the template | Errors found when"
                    " processing a rule on the template |  |"
                    " [Source](https://github.com/aws-cloudformation/cfn-lint) |"
                    " `base`,`rule` |\n"
                ),
                call(
                    '| [E0003<a name="E0003"></a>]'
                    "(../src/cfnlint/rules/errors/config.py) | "
                    "Error with cfn-lint configuration | "
                    "Error as a result of the cfn-lint configuration |  | "
                    "[Source](https://github.com/aws-cloudformation/cfn-lint) "
                    "| `base`,`rule` |\n"
                ),
                call("\n\\* experimental rules\n"),
            ]
            mock_builtin_open.return_value.write.assert_has_calls(expected_calls)
            self.assertEqual(
                len(expected_calls), mock_builtin_open.return_value.write.call_count
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from cfnlint.rules import CloudFormationLintRule


class IfUnsatisfiable(CloudFormationLintRule):
    id = "W1028"
    shortdesc = "Check Fn::If has a path that cannot be reached"
    description = "Check Fn::If path can be reached"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-if"
    tags = ["functions", "if"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.formatters.base import BaseFormatter


class Formatter(BaseFormatter):
    """Generic Formatter"""

    def _format(self, match):
        """Format output"""
        formatstr = "{0} {1}\n{2}:{3}:{4}\n"
        return formatstr.format(
            match.rule.id,
            match.message,
            match.filename,
            match.linenumber,
            match.columnnumber,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.formats._schema_comparer import compare_schemas


@pytest.mark.parametrize(
    "name,source,destination,expected",
    [
        (
            "basic valid",
            {"format": "foo"},
            {"format": "foo"},
            None,
        ),
        (
            "basic invalid",
            {"format": "foo"},
            {"format": "bar"},
            ValidationError(
                "'foo' format is incompatible with formats ['bar']",
            ),
        ),
        (
            "basic invalid with no destination",
            {"format": "foo"},
            {},
            ValidationError(
                "'foo' format is incompatible",
            ),
        ),
        (
            "backwards compatibility",
            {"format": "AWS::EC2::SecurityGroup.GroupId"},
            {"format": "AWS::EC2::SecurityGroup.Id"},
            None,
        ),
        (
            "basic valid with anyOf source",
            {"anyOf": [{"format": "foo"}, {"format": "bar"}]},
            {"format": "foo"},
            None,
        ),
        (
            "basic valid with anyOf source with no valid formats",
            {"anyOf": [{"format": "bar"}, {"format": "foobar"}]},
            {"format": "foo"},
            None,
        ),
        (
            "basic valid with anyOf destination",
            {"format": "foo"},
            {"anyOf": [{"format": "foo"}, {"format": "bar"}]},
            None,
        ),
        (
            "basic invalid with anyOf destination with many invalid",
            {"format": "foo"},
            {"anyOf": [{"format": "bar"}, {"format": "foobar"}]},
            ValidationError(
                "'foo' format is incompatible with formats ['bar', 'foobar']",
            ),
        ),
    ],
)
def test_schema_comparer(name, source, destination, expected):
    result = compare_schemas(source, destination)

    assert result == expected, f"{name!r} got errors {result!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class EventSourceMappingToSqsTimeout(CfnLintKeyword):

    id = "E3505"
    shortdesc = (
        "Validate SQS 'VisibilityTimeout' is greater than a function's 'Timeout'"
    )
    description = (
        "When attaching a Lambda function to a SQS queue to a Lambda function the "
        "SQS 'VisibilityTimeout' has to be greater than or equal to "
        " the lambda functions's 'Timeout'"
    )
    source_url = "https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-visibility-timeout.html"
    tags = ["resources", "lambda", "sqs"]

    def __init__(self):
        """Init"""
        super().__init__(["Resources/AWS::Lambda::Function/Properties/Timeout"])

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        if validator.is_type(instance, "string"):
            try:
                instance = int(instance)
            except:  # noqa: E722
                return

        if not validator.is_type(instance, "integer"):
            return

        if validator.cfn.graph is None:  # pragma: no cover
            return  # pragma: no cover

        if not len(validator.context.path.path) >= 2:
            return

        resource_name = validator.context.path.path[1]
        for child_1, _ in validator.cfn.graph.graph.in_edges(resource_name):
            if child_1 not in validator.context.resources:
                continue

            if (
                validator.context.resources[child_1].type
                == "AWS::Lambda::EventSourceMapping"
            ):
                for _, child_2 in validator.cfn.graph.graph.out_edges(child_1):
                    if child_2 not in validator.context.resources:
                        continue
                    if validator.context.resources[child_2].type == "AWS::SQS::Queue":
                        for visibility_timeout, _ in get_value_from_path(
                            validator,
                            validator.cfn.template,
                            deque(
                                [
                                    "Resources",
                                    child_2,
                                    "Properties",
                                    "VisibilityTimeout",
                                ]
                            ),
                        ):
                            if validator.is_type(visibility_timeout, "string"):
                                try:
                                    visibility_timeout = int(visibility_timeout)
                                except:  # noqa: E722
                                    continue

                            if not validator.is_type(visibility_timeout, "integer"):
                                continue

                            if visibility_timeout < instance:
                                yield ValidationError(
                                    message=(
                                        f"Queue visibility timeout "
                                        f"({visibility_timeout!r}) "
                                        "is less than Function timeout "
                                        f"({instance!r}) seconds"
                                    ),
                                    rule=self,
                                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from copy import deepcopy

from cfnlint.decode.mark import Mark

LOGGER = logging.getLogger(__name__)


def create_str_node_class(cls):
    """
    Create string node class
    """

    class node_class(cls):
        """Node class created based on the input class"""

        def __init__(
            self, x, start_mark: Mark | None = None, end_mark: Mark | None = None
        ):
            try:
                cls.__init__(self, x)
            except TypeError:
                cls.__init__(self)

            self.start_mark = start_mark or Mark()
            self.end_mark = end_mark or Mark()

        # pylint: disable=bad-classmethod-argument, unused-argument
        def __new__(self, x, start_mark, end_mark):
            return cls.__new__(self, x)

        def __deepcopy__(self, memo):
            result = str_node(self, self.start_mark, self.end_mark)
            memo[id(self)] = result
            return result

    node_class.__name__ = f"{cls.__name__}_node"
    return node_class


def create_dict_node_class(cls):
    """
    Create dynamic node class
    """

    class node_class(cls):
        """Node class created based on the input class"""

        def __init__(
            self,
            x,
            start_mark: Mark | None = None,
            end_mark: Mark | None = None,
            using_merge: bool = False,
        ):
            try:
                cls.__init__(self, x)
            except TypeError:
                cls.__init__(self)
            self.start_mark = start_mark or Mark()
            self.end_mark = end_mark or Mark()
            self.using_merge = using_merge

        def __deepcopy__(self, memo):
            result = dict_node(self, self.start_mark, self.end_mark)
            memo[id(self)] = result
            for k, v in self.items():
                result[deepcopy(k)] = deepcopy(v, memo)

            return result

        def get(self, key, default=None):
            """Override the default get"""
            if isinstance(default, dict):
                default = dict_node(default, self.start_mark, self.end_mark)
            return super().get(key, default)

    node_class.__name__ = f"{cls.__name__}_node"
    return node_class


def create_dict_list_class(cls):
    """
    Create dynamic list class
    """

    class node_class(cls):
        """Node class created based on the input class"""

        def __init__(
            self, x, start_mark: Mark | None = None, end_mark: Mark | None = None
        ):
            try:
                cls.__init__(self, x)
            except TypeError:
                cls.__init__(self)
            self.start_mark = start_mark or Mark()
            self.end_mark = end_mark or Mark()

        def __deepcopy__(self, memo):
            result = list_node([], self.start_mark, self.end_mark)
            memo[id(self)] = result
            for _, v in enumerate(self):
                result.append(deepcopy(v, memo))

            return result

    node_class.__name__ = f"{cls.__name__}_node"
    return node_class


str_node = create_str_node_class(str)
dict_node = create_dict_node_class(dict)
list_node = create_dict_list_class(list)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import unittest

from cfnlint.jsonschema._format import cfn_format_checker
from cfnlint.jsonschema.exceptions import FormatError


class TestFormat(unittest.TestCase):
    def test_format_checker(self):
        self.assertEqual(
            str(cfn_format_checker),
            (
                "<FormatChecker checkers=['date', 'date-time',"
                " 'email', 'ipv4', 'ipv4-network', 'ipv6', "
                "'ipv6-network', 'regex', 'time']>"
            ),
        )

        self.assertIsNone(cfn_format_checker.check("aws@amazon.com", "email"))
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "email"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("foo", "email")

        self.assertIsNone(cfn_format_checker.check("127.0.0.1", "ipv4"))
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "ipv4"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("foo", "ipv4")

        self.assertIsNone(cfn_format_checker.check("10.0.0.0/24", "ipv4-network"))
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "ipv4-network"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("foo", "ipv4-network")

        self.assertIsNone(
            cfn_format_checker.check("0000:0000:0000:0000:0000:0000:0000:0000", "ipv6")
        )
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "ipv6"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("foo", "ipv6")

        self.assertIsNone(cfn_format_checker.check("fd00:10:0:2::/64", "ipv6-network"))
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "ipv6-network"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("foo", "ipv6-network")

        self.assertIsNone(cfn_format_checker.check("1970-01-01", "date"))
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "date"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("foo", "date")

        self.assertIsNone(
            cfn_format_checker.check("1970-01-01T00:00:00.000Z", "date-time")
        )
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "date-time"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("foo", "date-time")

        self.assertIsNone(cfn_format_checker.check("00:00:00.000Z", "time"))
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "time"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("foo", "time")

        self.assertIsNone(cfn_format_checker.check("^$", "regex"))
        self.assertIsNone(cfn_format_checker.check({"foo": "bar"}, "regex"))
        with self.assertRaises(FormatError):
            cfn_format_checker.check("[a-", "regex")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import regex as re

from cfnlint.data import AdditionalSpecs
from cfnlint.helpers import ensure_list, load_resource
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class Permissions(CfnLintKeyword):
    """Check IAM Permission configuration"""

    id = "W3037"
    shortdesc = "Check IAM Permission configuration"
    description = "Check for valid IAM Permissions"
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_action.html"
    tags = ["properties", "iam", "permissions"]

    def __init__(self):
        """Init"""
        super().__init__(
            ["AWS::IAM::Policy/Properties/PolicyDocument/Statement/Action"]
        )
        self._service_map = load_resource(AdditionalSpecs, "Policies.json")
        self._resource_action_limitations = {
            "AWS::S3::BucketPolicy": ["s3"],
            "AWS::SQS::QueuePolicy": ["sqs"],
            "AWS::SNS::TopicPolicy": ["sns"],
        }

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        # Escape validation when using SAM transforms as a result of
        # https://github.com/aws/serverless-application-model/issues/3633
        if validator.cfn.has_serverless_transform():
            return

        actions = ensure_list(instance)

        for action in actions:
            if not validator.is_type(action, "string"):
                continue
            if action == "*":
                continue
            if ":" not in action:
                yield ValidationError(
                    (
                        f"{action!r} is not a valid action. "
                        "Must be of the form service:action or '*'"
                    ),
                    rule=self,
                )
                return
            service, permission = action.split(":", 1)
            service = service.lower()
            permission = permission.lower()

            if len(validator.context.path.cfn_path) >= 2:
                if (
                    validator.context.path.cfn_path[1]
                    in self._resource_action_limitations
                ):
                    if (
                        service
                        not in self._resource_action_limitations[
                            validator.context.path.cfn_path[1]
                        ]
                    ):
                        yield ValidationError(
                            (
                                f"{service!r} is not one of "
                                f"{self._resource_action_limitations[validator.context.path.cfn_path[1]]!r}"
                            ),
                            rule=self,
                        )

            if service in self._service_map:
                enums = list(self._service_map[service].get("Actions", []).keys())
                if permission == "*":
                    pass

                if any(x in permission for x in ["*", "?"]):
                    permission_regex = (
                        f"^{permission.replace('*', '.*').replace('?', '.')}$"
                    )
                    if not any(re.match(permission_regex, action) for action in enums):
                        yield ValidationError(
                            f"{permission!r} is not one of {enums!r}",
                            rule=self,
                        )
                elif permission not in enums:
                    yield ValidationError(
                        f"{permission!r} is not one of {enums!r}",
                        rule=self,
                    )
            else:
                yield ValidationError(
                    f"{service!r} is not one of {list(self._service_map.keys())!r}",
                    rule=self,
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import os
from typing import Sequence

from cfnlint.config import _DEFAULT_RULESDIR, ConfigMixIn, ManualArgs
from cfnlint.exceptions import UnexpectedRuleException
from cfnlint.match import Match
from cfnlint.rules import RulesCollection
from cfnlint.runner import TemplateRunner


def get_rules(
    append_rules: list[str],
    ignore_rules: list[str],
    include_rules: list[str],
    configure_rules=None,
    include_experimental: bool = False,
    mandatory_rules: list[str] | None = None,
    custom_rules: str | None = None,
) -> RulesCollection:
    rules = RulesCollection(
        ignore_rules,
        include_rules,
        configure_rules,
        include_experimental,
        mandatory_rules,
    )
    rules_paths: list[str] = [_DEFAULT_RULESDIR] + append_rules
    try:
        for rules_path in rules_paths:
            if rules_path and os.path.isdir(os.path.expanduser(rules_path)):
                rules.create_from_directory(rules_path)
            else:
                rules.create_from_module(rules_path)

        rules.create_from_custom_rules_file(custom_rules)
    except (OSError, ImportError) as e:
        raise UnexpectedRuleException(
            f"Tried to append rules but got an error: {str(e)}", 1
        ) from e
    return rules


def run_checks(
    filename: str,
    template: dict,
    rules: RulesCollection | None,
    regions: Sequence[str],
    mandatory_rules: Sequence[str] | None = None,
) -> list[Match]:
    """Run Checks and Custom Rules against the template"""

    config: ManualArgs = ManualArgs(
        regions=list(regions),
        mandatory_checks=list(mandatory_rules or []),
    )

    if not rules:
        return []

    config_mixin = ConfigMixIn(
        [],
        **config,
    )

    runner = TemplateRunner(
        filename=filename,
        template=template,
        rules=rules,  # type: ignore
        config=config_mixin,
    )

    return list(runner.run())



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from datetime import datetime
from typing import Any, Iterator, Tuple

import cfnlint.helpers
import cfnlint.rules.custom
from cfnlint._typing import Path, RuleMatches
from cfnlint.match import Match
from cfnlint.template import Template

LOGGER = logging.getLogger(__name__)


def _rule_is_enabled(
    rule: "CloudFormationLintRule",
    include_experimental: bool = False,
    ignore_rules: list[str] | None = None,
    include_rules: list[str] | None = None,
    mandatory_rules: list[str] | None = None,
):
    """Is the rule enabled based on the configuration"""
    ignore_rules = ignore_rules or []
    include_rules = include_rules or []
    mandatory_rules = mandatory_rules or []

    # Evaluate experimental rules
    if rule.experimental and not include_experimental:
        return False

    # Evaluate includes first:
    include_filter = False
    for include_rule in include_rules:
        if rule.id.startswith(include_rule):
            include_filter = True
    if not include_filter:
        return False

    # Enable mandatory rules without checking for if they are ignored
    for mandatory_rule in mandatory_rules:
        if rule.id.startswith(mandatory_rule):
            return True

    # Allowing ignoring of rules based on prefix to ignore checks
    for ignore_rule in ignore_rules:
        if rule.id.startswith(ignore_rule) and ignore_rule:
            return False

    return True


class RuleMatch:
    """
    Represents a rule match found by a CloudFormationLintRule.

    Attributes:
        path (Sequence[str | int]): The path to the element that
        triggered the rule match.
        path_string (str): The string representation of the path.
        message (str): The message associated with the rule match.
        context (RuleMatches): Additional context information
        related to the rule match.

    Methods:
        __eq__(self, item) -> bool:
            Override the equality comparison operator to compare
            rule matches based on their path and message.
        __hash__(self) -> int:
            Override the hash function to allow rule matches to
            be used as keys in a dictionary.
    """

    def __init__(self, path: Path, message: str, **kwargs):
        """
        Initialize a new RuleMatch instance.

        Args:
            path (Path): The path to the element
            that triggered the rule match.
            message (str): The message associated with the rule match.
            **kwargs: Additional keyword arguments to be stored
            as attributes on the RuleMatch instance.
        """
        self.path: Path = path
        self.path_string: str = "/".join(map(str, path))
        self.message: str = message
        self.context: RuleMatches = []
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __eq__(self, item):
        """
        Override the equality comparison operator to compare rule
        matches based on their path and message.

        Args:
            item (RuleMatch): The other RuleMatch instance to compare with.

        Returns:
            bool: True if the path and message of the two rule matches
            are equal, False otherwise.
        """
        return (self.path, self.message) == (item.path, item.message)

    def __hash__(self):
        """
        Override the hash function to allow rule matches to be
        used as keys in a dictionary.

        Returns:
            int: The hash value of the RuleMatch instance.
        """
        return hash((self.path, self.message))


def _rule_match_to_match(
    rule: CloudFormationLintRule,
    filename: str,
    cfn: Template,
    match: RuleMatch,
    parent_id: str | None = None,
) -> Iterator[Match]:
    error_rule = rule
    if hasattr(match, "rule"):
        error_rule = match.rule
    linenumbers: Tuple[int, int, int, int] | None = None
    if hasattr(match, "location"):
        linenumbers = match.location
    else:
        linenumbers = cfn.get_location_yaml(cfn.template, match.path)
    if linenumbers:
        result = Match.create(
            linenumber=linenumbers[0] + 1,
            columnnumber=linenumbers[1] + 1,
            linenumberend=linenumbers[2] + 1,
            columnnumberend=linenumbers[3] + 1,
            filename=filename,
            rule=error_rule,
            message=match.message,
            rulematch_obj=match,
            parent_id=parent_id,
        )
    else:
        result = Match.create(
            filename=filename,
            rule=error_rule,
            message=match.message,
            rulematch_obj=match,
            parent_id=parent_id,
        )

    yield result

    for sub_match in match.context:
        sub_match.path = list(match.path) + list(sub_match.path)
        yield from _rule_match_to_match(
            rule=rule, filename=filename, cfn=cfn, match=sub_match, parent_id=result.id
        )


def matching(match_type: Any):
    """Does Logging for match functions"""

    def decorator(match_function):
        """The Actual Decorator"""

        def wrapper(self, filename: str, cfn: Template, *args, **kwargs):
            """Wrapper"""
            if match_type == "match_resource_properties":
                if args[1] not in self.resource_property_types:
                    return

            start = datetime.now()
            LOGGER.debug("Starting match function for rule %s at %s", self.id, start)
            # pylint: disable=E1102
            for result in match_function(self, filename, cfn, *args, **kwargs):
                yield from _rule_match_to_match(self, filename, cfn, result)

            LOGGER.debug(
                "Complete match function for rule %s at %s.  Ran in %s",
                self.id,
                datetime.now(),
                datetime.now() - start,
            )

        return wrapper

    return decorator


class CloudFormationLintRule:
    """CloudFormation linter rules"""

    id: str = ""
    shortdesc: str = ""
    description: str = ""
    source_url: str = ""
    tags: list[str] = []
    experimental: bool = False

    logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        self.resource_property_types: list[str] = []
        self.config: dict[str, Any] = {}  # `-X E3012:strict=false`... Show more
        self.config_definition: dict[str, Any] = {}
        self._child_rules: dict[str, "CloudFormationLintRule" | None] = {}
        # Parent IDs to do the opposite of child rules
        self._parent_rules: list[str] = []
        super().__init__()

    def __repr__(self):
        return f"{self.id}: {self.shortdesc}"

    def __eq__(self, other):
        if other is None:
            return False
        return self.id == other.id

    @property
    def child_rules(self) -> dict[str, "CloudFormationLintRule" | None]:
        return self._child_rules

    @child_rules.setter
    def child_rules(self, rules: dict[str, "CloudFormationLintRule" | None]):
        self._child_rules = rules

    @property
    def parent_rules(self):
        return self._parent_rules

    @parent_rules.setter
    def parent_rules(self, rules: list[str]):
        self._parent_rules = rules

    @property
    def severity(self):
        """Severity level"""
        levels = {
            "I": "informational",
            "E": "error",
            "W": "warning",
        }
        return levels.get(self.id[0].upper(), "unknown")

    def verbose(self):
        """Verbose output"""
        return f"{self.id}: {self.shortdesc}\n{self.description}"

    def initialize(self, cfn: Template):
        """Initialize the rule"""

    def is_enabled(
        self,
        include_experimental=False,
        ignore_rules=None,
        include_rules=None,
        mandatory_rules=None,
    ):
        return _rule_is_enabled(
            self, include_experimental, ignore_rules, include_rules, mandatory_rules
        )

    def configure(self, configs=None, experimental=False):
        """Set the configuration"""

        # set defaults
        if isinstance(self.config_definition, dict):
            for config_name, config_values in self.config_definition.items():
                self.config[config_name] = config_values["default"]

        # set experimental if the rule is asking for it
        if "experimental" in self.config_definition:
            if self.config_definition["experimental"]["type"] == "boolean":
                self.config["experimental"] = cfnlint.helpers.bool_compare(
                    experimental, True
                )

        if isinstance(configs, dict):
            for key, value in configs.items():
                if key in self.config_definition:
                    if self.config_definition[key]["type"] == "boolean":
                        self.config[key] = cfnlint.helpers.bool_compare(value, True)
                    elif self.config_definition[key]["type"] == "string":
                        self.config[key] = str(value)
                    elif self.config_definition[key]["type"] == "integer":
                        self.config[key] = int(value)
                    elif self.config_definition[key]["type"] == "list":
                        self.config[key] = []
                        for l_value in value:
                            if self.config_definition[key]["itemtype"] == "boolean":
                                self.config[key].append(
                                    cfnlint.helpers.bool_compare(l_value, True)
                                )
                            elif self.config_definition[key]["itemtype"] == "string":
                                self.config[key].append(str(l_value))
                            elif self.config_definition[key]["itemtype"] == "integer":
                                self.config[key].append(int(l_value))
                            else:
                                self.config[key].append(l_value)

    def match(self, cfn: Template) -> RuleMatches:
        return []

    def match_resource_properties(
        self,
        properties: dict[str, Any],
        resourcetype: str,
        path: Path,
        cfn: Template,
    ) -> RuleMatches:
        return []

    @matching("match")
    # pylint: disable=W0613
    def matchall(self, filename: str, cfn: Template):
        """Match the entire file"""
        return self.match(cfn)  # pylint: disable=E1102

    @matching("match_resource_properties")
    # pylint: disable=W0613
    def matchall_resource_properties(
        self, filename, cfn, resource_properties, property_type, path
    ):
        """Check for resource properties type"""
        return self.match_resource_properties(  # pylint: disable=E1102
            resource_properties, property_type, path, cfn
        )
